@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2019 Red Hat Inc.
+ * Copyright (C) 2017-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -60,6 +60,10 @@ struct sparse_array *alloc_sparse_array (bool debug);
 /* Free sparse array. */
 extern void free_sparse_array (struct sparse_array *sa);
 
+#define CLEANUP_FREE_SPARSE_ARRAY \
+  __attribute__((cleanup (cleanup_free_sparse_array)))
+extern void cleanup_free_sparse_array (struct sparse_array **sa);
+
 /* Read bytes from the sparse array.
  * Note this can never return an error and never allocates.
  */
@@ -74,6 +78,13 @@ extern int sparse_array_write (struct sparse_array *sa, const void *buf,
                                uint32_t count, uint64_t offset)
   __attribute__((__nonnull__ (1, 2)));
 
+/* Fill range with a single byte 'c'.  If c == '\0' then this is the
+ * same as sparse_array_zero below.
+ */
+extern int sparse_array_fill (struct sparse_array *sa, char c,
+                              uint32_t count, uint64_t offset)
+  __attribute__((__nonnull__ (1)));
+
 /* Zero byte range in the sparse array.
  *
  * Zeroing and trimming are the same operation (this implementation
@@ -91,4 +102,17 @@ extern int sparse_array_extents (struct sparse_array *sa,
                                  uint32_t count, uint64_t offset,
                                  struct nbdkit_extents *extents);
 
+/* Blit (copy) between two sparse arrays.
+ *
+ * Copy count bytes from (sa1, offset1) to (sa2, offset2).
+ *
+ * You may call this with sa1 == sa2, but overlapping ranges are
+ * undefined.
+ */
+extern int sparse_array_blit (struct sparse_array *sa1,
+                              struct sparse_array *sa2,
+                              uint32_t count,
+                              uint64_t offset1, uint64_t offset2)
+  __attribute__((__nonnull__ (1, 2)));
+
 #endif /* NBDKIT_SPARSE_H */
