@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018-2019 Red Hat Inc.
+ * Copyright (C) 2018-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -86,16 +86,23 @@ log_config (nbdkit_next_config *next, void *nxdata,
   return next (nxdata, key, value);
 }
 
-/* Open the logfile. */
 static int
 log_config_complete (nbdkit_next_config_complete *next, void *nxdata)
 {
-  int fd;
-
   if (!logfilename) {
     nbdkit_error ("missing logfile= parameter for the log filter");
     return -1;
   }
+
+  return next (nxdata);
+}
+
+/* Open the logfile. */
+static int
+log_get_ready (nbdkit_next_get_ready *next, void *nxdata)
+{
+  int fd;
+
   /* Using fopen("ae"/"we") would be more convenient, but as Haiku
    * still lacks that, use this instead. Atomicity is not essential
    * here since .config completes before threads that might fork, if
@@ -176,7 +183,7 @@ output_return (struct handle *h, const char *act, uint64_t id, int r, int *err)
 {
   const char *s = "Other=>EINVAL";
 
-  /* Only decode what connections.c:nbd_errno() recognizes */
+  /* Only decode what protocol.c:nbd_errno() recognizes */
   if (r == -1) {
     switch (*err) {
     case EROFS:
@@ -207,6 +214,15 @@ output_return (struct handle *h, const char *act, uint64_t id, int r, int *err)
       s = "ESHUTDOWN";
       break;
 #endif
+    case ENOTSUP:
+#if ENOTSUP != EOPNOTSUPP
+    case EOPNOTSUPP:
+#endif
+      s = "ENOTSUP";
+      break;
+    case EOVERFLOW:
+      s = "EOVERFLOW";
+      break;
     case EINVAL:
       s = "EINVAL";
       break;
@@ -444,6 +460,7 @@ static struct nbdkit_filter filter = {
   .config_complete   = log_config_complete,
   .config_help       = log_config_help,
   .unload            = log_unload,
+  .get_ready         = log_get_ready,
   .open              = log_open,
   .close             = log_close,
   .prepare           = log_prepare,
