@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -134,6 +134,10 @@ struct nbdkit_plugin {
   int (*thread_model) (void);
 
   int (*can_fast_zero) (void *handle);
+
+  int (*preconnect) (int readonly);
+
+  int (*get_ready) (void);
 };
 
 extern void nbdkit_set_error (int err);
