@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2014 Red Hat Inc.
+ * Copyright (C) 2014-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -30,14 +30,6 @@
  * SUCH DAMAGE.
  */
 
-/* NB: The terminology used by libcurl is confusing!
- *
- * WRITEFUNCTION / write_cb is used when reading from the remote server
- * READFUNCTION / read_cb is used when writing to the remote server.
- *
- * We use the same terminology as libcurl here.
- */
-
 #include <config.h>
 
 #include <stdio.h>
@@ -56,23 +48,38 @@
 
 #include <nbdkit-plugin.h>
 
-static const char *url = NULL;
-static const char *user = NULL;
-static char *password = NULL;
-static const char *proxy_user = NULL;
-static char *proxy_password = NULL;
-static char *cookie = NULL;
-static bool sslverify = true;
-static uint32_t timeout = 0;
-static const char *unix_socket_path = NULL;
-static long protocols = CURLPROTO_ALL;
-
-/* Use '-D curl.verbose=1' to set. */
-int curl_debug_verbose = 0;
+#include "cleanup.h"
+#include "ascii-ctype.h"
+#include "ascii-string.h"
 
+/* Macro CURL_AT_LEAST_VERSION was added in 2015 (Curl 7.43) so if the
+ * macro isn't present then Curl is very old.
+ */
+#ifdef CURL_AT_LEAST_VERSION
 #if CURL_AT_LEAST_VERSION(7, 55, 0)
 #define HAVE_CURLINFO_CONTENT_LENGTH_DOWNLOAD_T
 #endif
+#endif
+
+static const char *url = NULL;  /* required */
+
+static const char *cainfo = NULL;
+static const char *capath = NULL;
+static char *cookie = NULL;
+static char *password = NULL;
+static long protocols = CURLPROTO_ALL;
+static const char *proxy = NULL;
+static char *proxy_password = NULL;
+static const char *proxy_user = NULL;
+static bool sslverify = true;
+static bool tcp_keepalive = false;
+static bool tcp_nodelay = true;
+static uint32_t timeout = 0;
+static const char *unix_socket_path = NULL;
+static const char *user = NULL;
+
+/* Use '-D curl.verbose=1' to set. */
+int curl_debug_verbose = 0;
 
 static void
 curl_load (void)
@@ -174,25 +181,12 @@ curl_config (const char *key, const char *value)
 {
   int r;
 
-  if (strcmp (key, "url") == 0)
-    url = value;
-
-  else if (strcmp (key, "user") == 0)
-    user = value;
-
-  else if (strcmp (key, "password") == 0) {
-    free (password);
-    if (nbdkit_read_password (value, &password) == -1)
-      return -1;
+  if (strcmp (key, "cainfo") == 0) {
+    cainfo = value;
   }
 
-  else if (strcmp (key, "proxy-user") == 0)
-    proxy_user = value;
-
-  else if (strcmp (key, "proxy-password") == 0) {
-    free (proxy_password);
-    if (nbdkit_read_password (value, &proxy_password) == -1)
-      return -1;
+  else if (strcmp (key, "capath") == 0) {
+    capath =  value;
   }
 
   else if (strcmp (key, "cookie") == 0) {
@@ -201,6 +195,30 @@ curl_config (const char *key, const char *value)
       return -1;
   }
 
+  else if (strcmp (key, "password") == 0) {
+    free (password);
+    if (nbdkit_read_password (value, &password) == -1)
+      return -1;
+  }
+
+  else if (strcmp (key, "protocols") == 0) {
+    if (parse_protocols (value) == -1)
+      return -1;
+  }
+
+  else if (strcmp (key, "proxy") == 0) {
+    proxy = value;
+  }
+
+  else if (strcmp (key, "proxy-password") == 0) {
+    free (proxy_password);
+    if (nbdkit_read_password (value, &proxy_password) == -1)
+      return -1;
+  }
+
+  else if (strcmp (key, "proxy-user") == 0)
+    proxy_user = value;
+
   else if (strcmp (key, "sslverify") == 0) {
     r = nbdkit_parse_bool (value);
     if (r == -1)
@@ -208,6 +226,20 @@ curl_config (const char *key, const char *value)
     sslverify = r;
   }
 
+  else if (strcmp (key, "tcp-keepalive") == 0) {
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    tcp_keepalive = r;
+  }
+
+  else if (strcmp (key, "tcp-nodelay") == 0) {
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    tcp_nodelay = r;
+  }
+
   else if (strcmp (key, "timeout") == 0) {
     if (nbdkit_parse_uint32_t ("timeout", value, &timeout) == -1)
       return -1;
@@ -226,10 +258,11 @@ curl_config (const char *key, const char *value)
            strcmp (key, "unix_socket_path") == 0)
     unix_socket_path = value;
 
-  else if (strcmp (key, "protocols") == 0) {
-    if (parse_protocols (value) == -1)
-      return -1;
-  }
+  else if (strcmp (key, "url") == 0)
+    url = value;
+
+  else if (strcmp (key, "user") == 0)
+    user = value;
 
   else {
     nbdkit_error ("unknown parameter '%s'", key);
@@ -253,13 +286,18 @@ curl_config_complete (void)
 }
 
 #define curl_config_help \
+  "cainfo=<CAINFO>            Path to Certificate Authority file.\n" \
+  "capath=<CAPATH>            Path to directory with CA certificates.\n" \
   "cookie=<COOKIE>            Set HTTP/HTTPS cookies.\n" \
   "password=<PASSWORD>        The password for the user account.\n" \
   "protocols=PROTO,PROTO,..   Limit protocols allowed.\n" \
+  "proxy=<PROXY>              Set proxy URL.\n" \
   "proxy-password=<PASSWORD>  The proxy password.\n" \
   "proxy-user=<USER>          The proxy user.\n" \
   "timeout=<TIMEOUT>          Set the timeout for requests (seconds).\n" \
   "sslverify=false            Do not verify SSL certificate of remote host.\n" \
+  "tcp-keepalive=true         Enable TCP keepalives.\n" \
+  "tcp-nodelay=false          Disable Nagle’s algorithm.\n" \
   "unix-socket-path=<PATH>    Open Unix domain socket instead of TCP/IP.\n" \
   "url=<URL>       (required) The disk image URL to serve.\n" \
   "user=<USER>                The user to log in as."
@@ -283,6 +321,8 @@ struct curl_handle {
                   curl_easy_strerror ((r)), (h)->errbuf);       \
   } while (0)
 
+static int debug_cb (CURL *handle, curl_infotype type,
+                     const char *data, size_t size, void *);
 static size_t header_cb (void *ptr, size_t size, size_t nmemb, void *opaque);
 static size_t write_cb (char *ptr, size_t size, size_t nmemb, void *opaque);
 static size_t read_cb (void *ptr, size_t size, size_t nmemb, void *opaque);
@@ -311,13 +351,13 @@ curl_open (int readonly)
     goto err;
   }
 
-  nbdkit_debug ("opened libcurl easy handle");
-
-  /* Note this writes the output to stderr directly.  We should
-   * consider using CURLOPT_DEBUGFUNCTION so we can handle it with
-   * nbdkit_debug.
-   */
-  curl_easy_setopt (h->c, CURLOPT_VERBOSE, curl_debug_verbose);
+  if (curl_debug_verbose) {
+    /* NB: Constants must be explicitly long because the parameter is
+     * varargs.
+     */
+    curl_easy_setopt (h->c, CURLOPT_VERBOSE, 1L);
+    curl_easy_setopt (h->c, CURLOPT_DEBUGFUNCTION, debug_cb);
+  }
 
   curl_easy_setopt (h->c, CURLOPT_ERRORBUFFER, h->errbuf);
 
@@ -334,41 +374,55 @@ curl_open (int readonly)
     goto err;
   }
 
+  /* Set the URL. */
   r = curl_easy_setopt (h->c, CURLOPT_URL, url);
   if (r != CURLE_OK) {
     display_curl_error (h, r, "curl_easy_setopt: CURLOPT_URL [%s]", url);
     goto err;
   }
 
-  nbdkit_debug ("set libcurl URL: %s", url);
-
+  /* Various options we always set. */
   curl_easy_setopt (h->c, CURLOPT_AUTOREFERER, 1);
   curl_easy_setopt (h->c, CURLOPT_FOLLOWLOCATION, 1);
   curl_easy_setopt (h->c, CURLOPT_FAILONERROR, 1);
+
+  /* Options. */
+  if (cainfo)
+    curl_easy_setopt (h->c, CURLOPT_CAINFO, cainfo);
+  if (capath)
+    curl_easy_setopt (h->c, CURLOPT_CAPATH, capath);
+  if (cookie)
+    curl_easy_setopt (h->c, CURLOPT_COOKIE, cookie);
+  if (password)
+    curl_easy_setopt (h->c, CURLOPT_PASSWORD, password);
   if (protocols != CURLPROTO_ALL) {
     curl_easy_setopt (h->c, CURLOPT_PROTOCOLS, protocols);
     curl_easy_setopt (h->c, CURLOPT_REDIR_PROTOCOLS, protocols);
   }
+  if (proxy)
+    curl_easy_setopt (h->c, CURLOPT_PROXY, proxy);
+  if (proxy_password)
+    curl_easy_setopt (h->c, CURLOPT_PROXYPASSWORD, proxy_password);
+  if (proxy_user)
+    curl_easy_setopt (h->c, CURLOPT_PROXYUSERNAME, proxy_user);
+  if (!sslverify) {
+    /* NB: Constants must be explicitly long because the parameter is
+     * varargs.
+     */
+    curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYPEER, 0L);
+    curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYHOST, 0L);
+  }
+  if (tcp_keepalive)
+    curl_easy_setopt (h->c, CURLOPT_TCP_KEEPALIVE, 1L);
+  if (!tcp_nodelay)
+    curl_easy_setopt (h->c, CURLOPT_TCP_NODELAY, 0L);
   if (timeout > 0)
     /* NB: The cast is required here because the parameter is varargs
      * treated as long, and not type safe.
      */
     curl_easy_setopt (h->c, CURLOPT_TIMEOUT, (long) timeout);
-  if (!sslverify) {
-    /* See comment above. */
-    curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYPEER, 0L);
-    curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYHOST, 0L);
-  }
   if (user)
     curl_easy_setopt (h->c, CURLOPT_USERNAME, user);
-  if (password)
-    curl_easy_setopt (h->c, CURLOPT_PASSWORD, password);
-  if (proxy_user)
-    curl_easy_setopt (h->c, CURLOPT_PROXYUSERNAME, proxy_user);
-  if (proxy_password)
-    curl_easy_setopt (h->c, CURLOPT_PROXYPASSWORD, proxy_password);
-  if (cookie)
-    curl_easy_setopt (h->c, CURLOPT_COOKIE, cookie);
 
   /* Get the file size and also whether the remote HTTP server
    * supports byte ranges.
@@ -416,8 +470,8 @@ curl_open (int readonly)
 #endif
   nbdkit_debug ("content length: %" PRIi64, h->exportsize);
 
-  if (strncasecmp (url, "http://", strlen ("http://")) == 0 ||
-      strncasecmp (url, "https://", strlen ("https://")) == 0) {
+  if (ascii_strncasecmp (url, "http://", strlen ("http://")) == 0 ||
+      ascii_strncasecmp (url, "https://", strlen ("https://")) == 0) {
     if (!h->accept_range) {
       nbdkit_error ("server does not support 'range' (byte range) requests");
       goto err;
@@ -436,8 +490,6 @@ curl_open (int readonly)
     curl_easy_setopt (h->c, CURLOPT_READDATA, h);
   }
 
-  nbdkit_debug ("returning new handle %p", h);
-
   return h;
 
  err:
@@ -447,29 +499,80 @@ curl_open (int readonly)
   return NULL;
 }
 
+/* When using CURLOPT_VERBOSE, this callback is used to redirect
+ * messages to nbdkit_debug (instead of stderr).
+ */
+static int
+debug_cb (CURL *handle, curl_infotype type,
+          const char *data, size_t size, void *opaque)
+{
+  size_t origsize = size;
+  CLEANUP_FREE char *str;
+
+  /* The data parameter passed is NOT \0-terminated, but also it may
+   * have \n or \r\n line endings.  The only sane way to deal with
+   * this is to copy the string.  (The data strings may also be
+   * multi-line, but we don't deal with that here).
+   */
+  str = malloc (size + 1);
+  if (str == NULL)
+    goto out;
+  memcpy (str, data, size);
+  str[size] = '\0';
+
+  while (size > 0 && (str[size-1] == '\n' || str[size-1] == '\r')) {
+    str[size-1] = '\0';
+    size--;
+  }
+
+  switch (type) {
+  case CURLINFO_TEXT:
+    nbdkit_debug ("%s", str);
+    break;
+  case CURLINFO_HEADER_IN:
+    nbdkit_debug ("S: %s", str);
+    break;
+  case CURLINFO_HEADER_OUT:
+    nbdkit_debug ("C: %s", str);
+    break;
+  default:
+    /* Assume everything else is binary data that we cannot print. */
+    nbdkit_debug ("<data with size=%zu>", origsize);
+  }
+
+ out:
+  return 0;
+}
+
 static size_t
 header_cb (void *ptr, size_t size, size_t nmemb, void *opaque)
 {
   struct curl_handle *h = opaque;
   size_t realsize = size * nmemb;
-  size_t len;
-  const char *accept_line = "Accept-Ranges: bytes";
-  const char *line = ptr;
+  const char *header = ptr;
+  const char *end = header + realsize;
+  const char *accept_ranges = "accept-ranges:";
+  const char *bytes = "bytes";
 
-  if (realsize >= strlen (accept_line) &&
-      strncmp (line, accept_line, strlen (accept_line)) == 0)
-    h->accept_range = true;
+  if (realsize >= strlen (accept_ranges) &&
+      ascii_strncasecmp (header, accept_ranges, strlen (accept_ranges)) == 0) {
+    const char *p = strchr (header, ':') + 1;
 
-  /* Useful to print the server headers when debugging.  However we
-   * must strip off trailing \r?\n from each line.
-   */
-  len = realsize;
-  if (len > 0 && line[len-1] == '\n')
-    len--;
-  if (len > 0 && line[len-1] == '\r')
-    len--;
-  if (len > 0)
-    nbdkit_debug ("S: %.*s", (int) len, line);
+    /* Skip whitespace between the header name and value. */
+    while (p < end && *p && ascii_isspace (*p))
+      p++;
+
+    if (end - p >= strlen (bytes)
+        && strncmp (p, bytes, strlen (bytes)) == 0) {
+      /* Check that there is nothing but whitespace after the value. */
+      p += strlen (bytes);
+      while (p < end && *p && ascii_isspace (*p))
+        p++;
+
+      if (p == end || !*p)
+        h->accept_range = true;
+    }
+  }
 
   return realsize;
 }
@@ -495,6 +598,14 @@ curl_get_size (void *handle)
   return h->exportsize;
 }
 
+/* NB: The terminology used by libcurl is confusing!
+ *
+ * WRITEFUNCTION / write_cb is used when reading from the remote server
+ * READFUNCTION / read_cb is used when writing to the remote server.
+ *
+ * We use the same terminology as libcurl here.
+ */
+
 /* Read data from the remote server. */
 static int
 curl_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
