@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018-2019 Red Hat Inc.
+ * Copyright (C) 2018-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -50,6 +50,7 @@
 
 #include "cleanup.h"
 #include "sparse.h"
+#include "format.h"
 
 /* If raw|base64|data parameter seen. */
 static int data_seen = 0;
@@ -88,7 +89,7 @@ data_unload (void)
 
 /* Parse the base64 parameter. */
 static int
-read_base64 (const char *value)
+read_base64 (const char *value, int64_t *size_ret)
 {
 #if defined(HAVE_GNUTLS) && defined(HAVE_GNUTLS_BASE64_DECODE2)
   gnutls_datum_t in, out;
@@ -105,6 +106,7 @@ read_base64 (const char *value)
   if (sparse_array_write (sa, out.data, out.size, 0) == -1)
     return -1;
   free (out.data);
+  *size_ret = out.size;
   return 0;
 #else
   nbdkit_error ("base64 is not supported in this build of the plugin");
@@ -112,147 +114,6 @@ read_base64 (const char *value)
 #endif
 }
 
-/* Store file at current offset in the sparse array, updating
- * the offset.
- */
-static int
-store_file (const char *filename, int64_t *offset)
-{
-  FILE *fp;
-  char buf[BUFSIZ];
-  size_t n;
-
-  fp = fopen (filename, "r");
-  if (fp == NULL) {
-    nbdkit_error ("%s: %m", filename);
-    return -1;
-  }
-
-  while (!feof (fp)) {
-    n = fread (buf, 1, BUFSIZ, fp);
-    if (n > 0) {
-      if (sparse_array_write (sa, buf, n, *offset) == -1) {
-        fclose (fp);
-        return -1;
-      }
-    }
-    if (ferror (fp)) {
-      nbdkit_error ("fread: %s: %m", filename);
-      fclose (fp);
-      return -1;
-    }
-    (*offset) += n;
-  }
-
-  if (fclose (fp) == EOF) {
-    nbdkit_error ("fclose: %s: %m", filename);
-    return -1;
-  }
-
-  return 0;
-}
-
-/* Parse the data parameter. */
-static int
-read_data (const char *value)
-{
-  int64_t offset = 0;
-  size_t i, len = strlen (value);
-
-  for (i = 0; i < len; ++i) {
-    int64_t j, k;
-    int n;
-    char c, cc[2];
-
-    if (sscanf (&value[i], " @%" SCNi64 "%n", &j, &n) == 1) {
-      if (j < 0) {
-        nbdkit_error ("data parameter @OFFSET must not be negative");
-        return -1;
-      }
-      i += n;
-      offset = j;
-    }
-    else if (sscanf (&value[i], " %" SCNi64 "*%" SCNi64 "%n",
-                     &j, &k, &n) == 2) {
-      if (j < 0 || j > 255) {
-        nbdkit_error ("data parameter BYTE must be in the range 0..255");
-        return -1;
-      }
-      if (k < 0) {
-        nbdkit_error ("data parameter *N must be >= 0");
-        return -1;
-      }
-      i += n;
-
-      c = j;
-      while (k > 0) {
-        if (sparse_array_write (sa, &c, 1, offset) == -1)
-          return -1;
-        offset++;
-        k--;
-      }
-      if (data_size < offset)
-        data_size = offset;
-    }
-    /* We need %1s for obscure reasons.  sscanf " <%n" can return 0
-     * if nothing is matched, not only if the '<' is matched.
-     */
-    else if (sscanf (&value[i], " <%1s%n", cc, &n) == 1) {
-      CLEANUP_FREE char *filename = NULL;
-      size_t flen;
-
-      i += n-1;
-
-      /* The filename follows next in the string. */
-      flen = strcspn (&value[i], " \t\n");
-      if (flen == 0) {
-        nbdkit_error ("data parameter <FILE not a filename");
-        return -1;
-      }
-      filename = strndup (&value[i], flen);
-      if (filename == NULL) {
-        nbdkit_error ("strndup: %m");
-        return -1;
-      }
-      i += len;
-
-      if (store_file (filename, &offset) == -1)
-        return -1;
-
-      if (data_size < offset)
-        data_size = offset;
-    }
-    else if (sscanf (&value[i], " %" SCNi64 "%n", &j, &n) == 1) {
-      if (j < 0 || j > 255) {
-        nbdkit_error ("data parameter BYTE must be in the range 0..255");
-        return -1;
-      }
-      i += n;
-
-      if (data_size < offset+1)
-        data_size = offset+1;
-
-      /* Store the byte. */
-      c = j;
-      if (sparse_array_write (sa, &c, 1, offset) == -1)
-        return -1;
-      offset++;
-    }
-    /* We have to have a rule to skip just whitespace so that
-     * whitespace is permitted at the end of the string.
-     */
-    else if (sscanf (&value[i], " %n", &n) == 0) {
-      i += n;
-    }
-    else {
-      nbdkit_error ("data parameter: parsing error at offset %zu", i);
-      return -1;
-    }
-  }
-
-  return 0;
-}
-
 static int
 data_config (const char *key, const char *value)
 {
@@ -279,11 +140,11 @@ data_config (const char *key, const char *value)
         return -1;
     }
     else if (strcmp (key, "base64") == 0) {
-      if (read_base64 (value) == -1)
+      if (read_base64 (value, &data_size) == -1)
         return -1;
     }
     else if (strcmp (key, "data") == 0) {
-      if (read_data (value) == -1)
+      if (read_data_format (value, sa, &data_size) == -1)
         return -1;
     }
     else
@@ -320,7 +181,7 @@ data_config_complete (void)
 
 #define data_config_help \
   "data|raw|base64=...     Specify disk data on the command line\n" \
-  "size=<SIZE>  (required) Size of the backing disk"
+  "size=<SIZE>             Size of the backing disk"
 
 /* Provide a way to detect if the base64 feature is supported. */
 static void
