@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -55,15 +55,11 @@
 #endif
 
 #define NBDKIT_API_VERSION 2
-
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
 #include "isaligned.h"
 
-#ifndef O_CLOEXEC
-#define O_CLOEXEC 0
-#endif
-
 #ifndef HAVE_FDATASYNC
 #define fdatasync fsync
 #endif
@@ -76,6 +72,12 @@ static pthread_mutex_t lseek_lock = PTHREAD_MUTEX_INITIALIZER;
 /* to enable: -D file.zero=1 */
 int file_debug_zero;
 
+static bool
+is_enotsup (int err)
+{
+  return err == ENOTSUP || err == EOPNOTSUPP;
+}
+
 static void
 file_unload (void)
 {
@@ -186,7 +188,16 @@ file_open (int readonly)
     return NULL;
   }
 
-  h->is_block_device = S_ISBLK (statbuf.st_mode);
+  if (S_ISBLK (statbuf.st_mode))
+    h->is_block_device = true;
+  else if (S_ISREG (statbuf.st_mode))
+    h->is_block_device = false;
+  else {
+    nbdkit_error ("file is not regular or block device: %s", filename);
+    close (h->fd);
+    free (h);
+    return NULL;
+  }
   h->sector_size = 4096;  /* Start with safe guess */
 
 #ifdef BLKSSZGET
@@ -250,12 +261,8 @@ file_get_size (void *handle)
   struct handle *h = handle;
 
   if (h->is_block_device) {
-    int64_t size;
-
-    pthread_mutex_lock (&lseek_lock);
-    size = block_device_size (h->fd);
-    pthread_mutex_unlock (&lseek_lock);
-    return size;
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lseek_lock);
+    return block_device_size (h->fd);
   } else {
     /* Regular file. */
     struct stat statbuf;
@@ -297,6 +304,20 @@ file_can_fua (void *handle)
   return NBDKIT_FUA_NATIVE;
 }
 
+static int
+file_can_cache (void *handle)
+{
+  /* Prefer posix_fadvise(), but letting nbdkit call .pread on our
+   * behalf also tends to work well for the local file system
+   * cache.
+   */
+#if HAVE_POSIX_FADVISE
+  return NBDKIT_FUA_NATIVE;
+#else
+  return NBDKIT_FUA_EMULATE;
+#endif
+}
+
 /* Flush the file to disk. */
 static int
 file_flush (void *handle, uint32_t flags)
@@ -392,7 +413,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
       goto out;
     }
 
-    if (errno != EOPNOTSUPP) {
+    if (!is_enotsup (errno)) {
       nbdkit_error ("zero: %m");
       return -1;
     }
@@ -411,7 +432,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
       goto out;
     }
 
-    if (errno != EOPNOTSUPP) {
+    if (!is_enotsup (errno)) {
       nbdkit_error ("zero: %m");
       return -1;
     }
@@ -436,14 +457,14 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         goto out;
       }
 
-      if (errno != EOPNOTSUPP) {
+      if (!is_enotsup (errno)) {
         nbdkit_error ("zero: %m");
         return -1;
       }
 
       h->can_fallocate = false;
     } else {
-      if (errno != EOPNOTSUPP) {
+      if (!is_enotsup (errno)) {
         nbdkit_error ("zero: %m");
         return -1;
       }
@@ -506,7 +527,7 @@ file_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
-      if (errno == EOPNOTSUPP)
+      if (is_enotsup (EOPNOTSUPP))
         h->can_punch_hole = false;
 
       nbdkit_debug ("ignoring failed fallocate during trim: %m");
@@ -532,9 +553,8 @@ file_can_extents (void *handle)
   /* A simple test to see whether SEEK_HOLE etc is likely to work on
    * the current filesystem.
    */
-  pthread_mutex_lock (&lseek_lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lseek_lock);
   r = lseek (h->fd, 0, SEEK_HOLE);
-  pthread_mutex_unlock (&lseek_lock);
   if (r == -1) {
     nbdkit_debug ("extents disabled: lseek: SEEK_HOLE: %m");
     return 0;
@@ -607,16 +627,30 @@ static int
 file_extents (void *handle, uint32_t count, uint64_t offset,
               uint32_t flags, struct nbdkit_extents *extents)
 {
-  int r;
-
-  pthread_mutex_lock (&lseek_lock);
-  r = do_extents (handle, count, offset, flags, extents);
-  pthread_mutex_unlock (&lseek_lock);
-
-  return r;
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lseek_lock);
+  return do_extents (handle, count, offset, flags, extents);
 }
 #endif /* SEEK_HOLE */
 
+#if HAVE_POSIX_FADVISE
+/* Caching. */
+static int
+file_cache (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
+{
+  struct handle *h = handle;
+  int r;
+
+  /* Cache is advisory, we don't care if this fails */
+  r = posix_fadvise (h->fd, offset, count, POSIX_FADV_WILLNEED);
+  if (r) {
+    errno = r;
+    nbdkit_error ("posix_fadvise: %m");
+    return -1;
+  }
+  return 0;
+}
+#endif /* HAVE_POSIX_FADVISE */
+
 static struct nbdkit_plugin plugin = {
   .name              = "file",
   .longname          = "nbdkit file plugin",
@@ -633,6 +667,7 @@ static struct nbdkit_plugin plugin = {
   .can_multi_conn    = file_can_multi_conn,
   .can_trim          = file_can_trim,
   .can_fua           = file_can_fua,
+  .can_cache         = file_can_cache,
   .pread             = file_pread,
   .pwrite            = file_pwrite,
   .flush             = file_flush,
@@ -641,6 +676,9 @@ static struct nbdkit_plugin plugin = {
 #ifdef SEEK_HOLE
   .can_extents       = file_can_extents,
   .extents           = file_extents,
+#endif
+#if HAVE_POSIX_FADVISE
+  .cache             = file_cache,
 #endif
   .errno_is_preserved = 1,
 };
