@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2019 Red Hat Inc.
+ * Copyright (C) 2017-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -43,12 +43,15 @@
 
 #include <pthread.h>
 
+#define NBDKIT_API_VERSION 2
+
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
 #include "sparse.h"
 
 /* The size of disk in bytes (initialized by size=<SIZE> parameter). */
-static int64_t size = 0;
+static int64_t size = -1;
 
 /* Debug directory operations (-D memory.dir=1). */
 int memory_debug_dir;
@@ -94,7 +97,7 @@ memory_config (const char *key, const char *value)
 static int
 memory_config_complete (void)
 {
-  if (size == 0) {
+  if (size == -1) {
     nbdkit_error ("you must specify size=<SIZE> on the command line");
     return -1;
   }
@@ -117,7 +120,14 @@ memory_open (int readonly)
 static int64_t
 memory_get_size (void *handle)
 {
-  return (int64_t) size;
+  return size;
+}
+
+/* Flush is a no-op, so advertise native FUA support */
+static int
+memory_can_fua (void *handle)
+{
+  return NBDKIT_FUA_NATIVE;
 }
 
 /* Serves the same data over multiple connections. */
@@ -127,45 +137,73 @@ memory_can_multi_conn (void *handle)
   return 1;
 }
 
+/* Cache. */
+static int
+memory_can_cache (void *handle)
+{
+  /* Everything is already in memory, returning this without
+   * implementing .cache lets nbdkit do the correct no-op.
+   */
+  return NBDKIT_CACHE_NATIVE;
+}
+
+/* Fast zero. */
+static int
+memory_can_fast_zero (void *handle)
+{
+  return 1;
+}
+
 /* Read data. */
 static int
-memory_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
+memory_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
+              uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  assert (!flags);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_read (sa, buf, count, offset);
-  pthread_mutex_unlock (&lock);
   return 0;
 }
 
 /* Write data. */
 static int
-memory_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset)
+memory_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
+               uint32_t flags)
 {
-  int r;
-
-  pthread_mutex_lock (&lock);
-  r = sparse_array_write (sa, buf, count, offset);
-  pthread_mutex_unlock (&lock);
-  return r;
+  /* Flushing, and thus FUA flag, is a no-op */
+  assert ((flags & ~NBDKIT_FLAG_FUA) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+  return sparse_array_write (sa, buf, count, offset);
 }
 
 /* Zero. */
 static int
-memory_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+memory_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  /* Flushing, and thus FUA flag, is a no-op. Assume that
+   * sparse_array_zero generally beats writes, so FAST_ZERO is a no-op. */
+  assert ((flags & ~(NBDKIT_FLAG_FUA | NBDKIT_FLAG_MAY_TRIM |
+                     NBDKIT_FLAG_FAST_ZERO)) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_zero (sa, count, offset);
-  pthread_mutex_unlock (&lock);
   return 0;
 }
 
 /* Trim (same as zero). */
 static int
-memory_trim (void *handle, uint32_t count, uint64_t offset)
+memory_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  /* Flushing, and thus FUA flag, is a no-op */
+  assert ((flags & ~NBDKIT_FLAG_FUA) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_zero (sa, count, offset);
-  pthread_mutex_unlock (&lock);
+  return 0;
+}
+
+/* Nothing is persistent, so flush is trivially supported */
+static int
+memory_flush (void *handle, uint32_t flags)
+{
   return 0;
 }
 
@@ -174,12 +212,8 @@ static int
 memory_extents (void *handle, uint32_t count, uint64_t offset,
                 uint32_t flags, struct nbdkit_extents *extents)
 {
-  int r;
-
-  pthread_mutex_lock (&lock);
-  r = sparse_array_extents (sa, count, offset, extents);
-  pthread_mutex_unlock (&lock);
-  return r;
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+  return sparse_array_extents (sa, count, offset, extents);
 }
 
 static struct nbdkit_plugin plugin = {
@@ -190,13 +224,18 @@ static struct nbdkit_plugin plugin = {
   .config            = memory_config,
   .config_complete   = memory_config_complete,
   .config_help       = memory_config_help,
+  .magic_config_key  = "size",
   .open              = memory_open,
   .get_size          = memory_get_size,
+  .can_fua           = memory_can_fua,
   .can_multi_conn    = memory_can_multi_conn,
+  .can_cache         = memory_can_cache,
+  .can_fast_zero     = memory_can_fast_zero,
   .pread             = memory_pread,
   .pwrite            = memory_pwrite,
   .zero              = memory_zero,
   .trim              = memory_trim,
+  .flush             = memory_flush,
   .extents           = memory_extents,
   /* In this plugin, errno is preserved properly along error return
    * paths from failed system calls.
