@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2019 Red Hat Inc.
+ * Copyright (C) 2017-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -52,6 +52,8 @@
 #define NBDKIT_API_VERSION 2
 
 #include <nbdkit-plugin.h>
+
+#include "ascii-string.h"
 #include "byte-swapping.h"
 #include "cleanup.h"
 #include "utils.h"
@@ -69,12 +71,14 @@ struct transaction {
 struct handle {
   /* These fields are read-only once initialized */
   struct nbd_handle *nbd;
-  int fd; /* Cache of nbd_aio_get_fd */
   int fds[2]; /* Pipe for kicking the reader thread */
   bool readonly;
   pthread_t reader;
 };
 
+/* Connect to server via URI */
+static const char *uri;
+
 /* Connect to server via absolute name of Unix socket */
 static char *sockname;
 
@@ -82,9 +86,6 @@ static char *sockname;
 static const char *hostname;
 static const char *port;
 
-/* Connect to server via URI */
-static const char *uri;
-
 /* Name of export on remote server, default '', ignored for oldstyle */
 static const char *export;
 
@@ -152,9 +153,9 @@ nbdplug_config (const char *key, const char *value)
     shared = r;
   }
   else if (strcmp (key, "tls") == 0) {
-    if (strcasecmp (value, "require") == 0 ||
-        strcasecmp (value, "required") == 0 ||
-        strcasecmp (value, "force") == 0)
+    if (ascii_strcasecmp (value, "require") == 0 ||
+        ascii_strcasecmp (value, "required") == 0 ||
+        ascii_strcasecmp (value, "force") == 0)
       tls = LIBNBD_TLS_REQUIRE;
     else {
       r = nbdkit_parse_bool (value);
@@ -191,41 +192,32 @@ nbdplug_config (const char *key, const char *value)
   return 0;
 }
 
-/* Check the user passed exactly one socket description. */
 static int
 nbdplug_config_complete (void)
 {
-  if (sockname) {
-    struct sockaddr_un sock;
+  int c = !!sockname + !!hostname + !!uri;
 
-    if (hostname || port) {
-      nbdkit_error ("cannot mix Unix socket and TCP hostname/port parameters");
-      return -1;
-    }
-    else if (uri) {
-      nbdkit_error ("cannot mix Unix socket and URI parameters");
-      return -1;
-    }
-    if (strlen (sockname) > sizeof sock.sun_path) {
-      nbdkit_error ("socket file name too large");
-      return -1;
-    }
+  /* Check the user passed exactly one connection parameter. */
+  if (c > 1) {
+    nbdkit_error ("cannot mix Unix ‘socket’, TCP ‘hostname’/‘port’ "
+                  "and ‘uri’ parameters");
+    return -1;
   }
-  else if (hostname) {
-    if (uri) {
-      nbdkit_error ("cannot mix TCP hostname/port and URI parameters");
-      return -1;
-    }
-    if (!port)
-      port = "10809";
+  if (c == 0) {
+    nbdkit_error ("exactly one of ‘socket’, ‘hostname’ "
+                  "and ‘uri’ parameters must be specified");
+    return -1;
   }
-  else if (uri) {
+
+  /* Port, if present, should only be used with hostname. */
+  if (port && !hostname) {
+    nbdkit_error ("‘port’ parameter should only be used with ‘hostname’");
+    return -1;
+  }
+
+  if (uri) {
     struct nbd_handle *nbd = nbd_create ();
 
-    if (port) {
-      nbdkit_error ("cannot mix TCP hostname/port and URI parameters");
-      return -1;
-    }
     if (!nbd) {
       nbdkit_error ("unable to query libnbd details: %s", nbd_get_error ());
       return -1;
@@ -236,11 +228,24 @@ nbdplug_config_complete (void)
       return -1;
     }
     nbd_close (nbd);
-  } else {
-    nbdkit_error ("must supply socket=, hostname= or uri= of external NBD server");
-    return -1;
+  }
+  else if (sockname) {
+    struct sockaddr_un sock;
+
+    if (strlen (sockname) > sizeof sock.sun_path) {
+      nbdkit_error ("socket file name too large");
+      return -1;
+    }
+  }
+  else if (hostname) {
+    if (!port)
+      port = "10809";
+  }
+  else {
+    abort ();         /* can't happen, if checks above were correct */
   }
 
+  /* Check the other parameters. */
   if (!export)
     export = "";
 
@@ -262,6 +267,7 @@ nbdplug_config_complete (void)
     nbd_close (nbd);
   }
 
+  /* Create the shared connection. */
   if (shared && (shared_handle = nbdplug_open_handle (false)) == NULL)
     return -1;
   return 0;
@@ -305,14 +311,16 @@ nbdplug_reader (void *handle)
 {
   struct handle *h = handle;
 
+  nbdkit_debug ("nbd: started reader thread");
+
   while (!nbd_aio_is_dead (h->nbd) && !nbd_aio_is_closed (h->nbd)) {
+    int r;
     struct pollfd fds[2] = {
-      [0].fd = h->fd,
+      [0].fd = nbd_aio_get_fd (h->nbd),
       [1].fd = h->fds[0],
       [1].events = POLLIN,
     };
     unsigned dir;
-    char c;
 
     dir = nbd_aio_get_direction (h->nbd);
     nbdkit_debug ("polling, dir=%d", dir);
@@ -325,16 +333,23 @@ nbdplug_reader (void *handle)
       break;
     }
 
-    if (dir & LIBNBD_AIO_DIRECTION_READ && fds[0].revents & POLLIN)
-      nbd_aio_notify_read (h->nbd);
-    else if (dir & LIBNBD_AIO_DIRECTION_WRITE && fds[0].revents & POLLOUT)
-      nbd_aio_notify_write (h->nbd);
+    dir = nbd_aio_get_direction (h->nbd);
+
+    r = 0;
+    if ((dir & LIBNBD_AIO_DIRECTION_READ) && (fds[0].revents & POLLIN))
+      r = nbd_aio_notify_read (h->nbd);
+    else if ((dir & LIBNBD_AIO_DIRECTION_WRITE) && (fds[0].revents & POLLOUT))
+      r = nbd_aio_notify_write (h->nbd);
+    if (r == -1) {
+      nbdkit_error ("%s", nbd_get_error ());
+      break;
+    }
 
     /* Check if we were kicked because a command was started */
     if (fds[1].revents & POLLIN) {
-      while (read (h->fds[0], &c, 1) == 1)
-        /* Drain any backlog */;
-      if (errno != EAGAIN) {
+      char buf[10]; /* Larger than 1 to allow reduction of any backlog */
+
+      if (read (h->fds[0], buf, sizeof buf) == -1 && errno != EAGAIN) {
         nbdkit_error ("failed to read pipe: %m");
         break;
       }
@@ -342,7 +357,7 @@ nbdplug_reader (void *handle)
   }
 
   nbdkit_debug ("state machine changed to %s", nbd_connection_state (h->nbd));
-  nbdkit_debug ("exiting state machine thread");
+  nbdkit_debug ("exiting reader thread");
   return NULL;
 }
 
@@ -392,7 +407,7 @@ nbdplug_register (struct handle *h, struct transaction *trans, int64_t cookie)
   nbdkit_debug ("cookie %" PRId64 " started by state machine", cookie);
   trans->cookie = cookie;
 
-  if (write (h->fds[1], &c, 1) != 1 && errno != EAGAIN)
+  if (write (h->fds[1], &c, 1) == -1 && errno != EAGAIN)
     nbdkit_debug ("failed to kick reader thread: %m");
 }
 
@@ -461,7 +476,6 @@ nbdplug_open_handle (int readonly)
 #endif
 
  retry:
-  h->fd = -1;
   h->nbd = nbd_create ();
   if (!h->nbd)
     goto err;
@@ -484,8 +498,10 @@ nbdplug_open_handle (int readonly)
     r = nbd_connect_uri (h->nbd, uri);
   else if (sockname)
     r = nbd_connect_unix (h->nbd, sockname);
-  else
+  else if (hostname)
     r = nbd_connect_tcp (h->nbd, hostname, port);
+  else
+    abort ();
   if (r == -1) {
     if (retries--) {
       nbdkit_debug ("connect failed; will try again: %s", nbd_get_error ());
@@ -495,9 +511,6 @@ nbdplug_open_handle (int readonly)
     }
     goto err;
   }
-  h->fd = nbd_aio_get_fd (h->nbd);
-  if (h->fd == -1)
-    goto err;
 
   if (readonly)
     h->readonly = true;
@@ -533,7 +546,7 @@ nbdplug_open (int readonly)
 static void
 nbdplug_close_handle (struct handle *h)
 {
-  if (nbd_shutdown (h->nbd, 0) == -1)
+  if (nbd_aio_disconnect (h->nbd, 0) == -1)
     nbdkit_debug ("failed to clean up handle: %s", nbd_get_error ());
   if ((errno = pthread_join (h->reader, NULL)))
     nbdkit_debug ("failed to join reader thread: %m");
