@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2014-2019 Red Hat Inc.
+ * Copyright (C) 2014-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -49,6 +49,19 @@
 
 #include <nbdkit-plugin.h>
 
+/* Replacement if caml_alloc_initialized_string is missing, added
+ * to OCaml runtime in 2017.
+ */
+#ifndef HAVE_CAML_ALLOC_INITIALIZED_STRING
+static inline value
+caml_alloc_initialized_string (mlsize_t len, const char *p)
+{
+  value sv = caml_alloc_string (len);
+  memcpy ((char *) String_val (sv), p, len);
+  return sv;
+}
+#endif
+
 /* This constructor runs when the plugin loads, and initializes the
  * OCaml runtime, and lets the plugin set up its callbacks.
  */
@@ -100,42 +113,39 @@ plugin_init (void)
 static value load_fn;
 static value unload_fn;
 
+static value dump_plugin_fn;
+
 static value config_fn;
 static value config_complete_fn;
+static value thread_model_fn;
 
+static value get_ready_fn;
+
+static value preconnect_fn;
 static value open_fn;
 static value close_fn;
 
 static value get_size_fn;
 
-static value can_write_fn;
+static value can_cache_fn;
+static value can_extents_fn;
+static value can_fast_zero_fn;
 static value can_flush_fn;
-static value is_rotational_fn;
-static value can_trim_fn;
-
-static value dump_plugin_fn;
-
-static value can_zero_fn;
 static value can_fua_fn;
+static value can_multi_conn_fn;
+static value can_trim_fn;
+static value can_write_fn;
+static value can_zero_fn;
+static value is_rotational_fn;
 
 static value pread_fn;
 static value pwrite_fn;
 static value flush_fn;
 static value trim_fn;
 static value zero_fn;
-
-static value can_multi_conn_fn;
-
-static value can_extents_fn;
 static value extents_fn;
-
-static value can_cache_fn;
 static value cache_fn;
 
-static value thread_model_fn;
-
-static value can_fast_zero_fn;
-
 /*----------------------------------------------------------------------*/
 /* Wrapper functions that translate calls from C (ie. nbdkit) to OCaml. */
 
@@ -168,6 +178,21 @@ unload_wrapper (void)
   remove_roots ();
 }
 
+static void
+dump_plugin_wrapper (void)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (dump_plugin_fn, Val_unit);
+  if (Is_exception_result (rv))
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+  caml_enter_blocking_section ();
+  CAMLreturn0;
+}
+
 static int
 config_wrapper (const char *key, const char *val)
 {
@@ -209,6 +234,63 @@ config_complete_wrapper (void)
   CAMLreturnT (int, 0);
 }
 
+static int
+thread_model_wrapper (void)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (thread_model_fn, Val_unit);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Int_val (rv));
+}
+
+static int
+get_ready_wrapper (void)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (get_ready_fn, Val_unit);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, 0);
+}
+
+static int
+preconnect_wrapper (int readonly)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (preconnect_fn, Val_bool (readonly));
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, 0);
+}
+
 static void *
 open_wrapper (int readonly)
 {
@@ -354,21 +436,6 @@ can_trim_wrapper (void *h)
   CAMLreturnT (int, Bool_val (rv));
 }
 
-static void
-dump_plugin_wrapper (void)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (dump_plugin_fn, Val_unit);
-  if (Is_exception_result (rv))
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-  caml_enter_blocking_section ();
-  CAMLreturn0;
-}
-
 static int
 can_zero_wrapper (void *h)
 {
@@ -407,6 +474,82 @@ can_fua_wrapper (void *h)
   CAMLreturnT (int, Int_val (rv));
 }
 
+static int
+can_fast_zero_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_fast_zero_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Bool_val (rv));
+}
+
+static int
+can_cache_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_cache_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Int_val (rv));
+}
+
+static int
+can_extents_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_extents_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Bool_val (rv));
+}
+
+static int
+can_multi_conn_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_multi_conn_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Bool_val (rv));
+}
+
 static value
 Val_flags (uint32_t flags)
 {
@@ -480,8 +623,7 @@ pwrite_wrapper (void *h, const void *buf, uint32_t count, uint64_t offset,
 
   caml_leave_blocking_section ();
 
-  strv = caml_alloc_string (count);
-  memcpy (&Byte_u (strv, 0), buf, count);
+  strv = caml_alloc_initialized_string (count, buf);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
@@ -563,44 +705,6 @@ zero_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
   CAMLreturnT (int, 0);
 }
 
-static int
-can_multi_conn_wrapper (void *h)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (can_multi_conn_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    caml_enter_blocking_section ();
-    CAMLreturnT (int, -1);
-  }
-
-  caml_enter_blocking_section ();
-  CAMLreturnT (int, Bool_val (rv));
-}
-
-static int
-can_extents_wrapper (void *h)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (can_extents_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    caml_enter_blocking_section ();
-    CAMLreturnT (int, -1);
-  }
-
-  caml_enter_blocking_section ();
-  CAMLreturnT (int, Bool_val (rv));
-}
-
 static int
 extents_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags,
                  struct nbdkit_extents *extents)
@@ -646,25 +750,6 @@ extents_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags,
   CAMLreturnT (int, 0);
 }
 
-static int
-can_cache_wrapper (void *h)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (can_cache_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    caml_enter_blocking_section ();
-    CAMLreturnT (int, -1);
-  }
-
-  caml_enter_blocking_section ();
-  CAMLreturnT (int, Int_val (rv));
-}
-
 static int
 cache_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
 {
@@ -688,44 +773,6 @@ cache_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
   CAMLreturnT (int, 0);
 }
 
-static int
-thread_model_wrapper (void)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (thread_model_fn, Val_unit);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    caml_enter_blocking_section ();
-    CAMLreturnT (int, -1);
-  }
-
-  caml_enter_blocking_section ();
-  CAMLreturnT (int, Int_val (rv));
-}
-
-static int
-can_fast_zero_wrapper (void *h)
-{
-  CAMLparam0 ();
-  CAMLlocal1 (rv);
-
-  caml_leave_blocking_section ();
-
-  rv = caml_callback_exn (can_fast_zero_fn, *(value *) h);
-  if (Is_exception_result (rv)) {
-    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
-    caml_enter_blocking_section ();
-    CAMLreturnT (int, -1);
-  }
-
-  caml_enter_blocking_section ();
-  CAMLreturnT (int, Bool_val (rv));
-}
-
 /*----------------------------------------------------------------------*/
 /* set_* functions called from OCaml code at load time to initialize
  * fields in the plugin struct.
@@ -779,9 +826,15 @@ ocaml_nbdkit_set_config_help (value helpv)
 SET(load)
 SET(unload)
 
+SET(dump_plugin)
+
 SET(config)
 SET(config_complete)
+SET(thread_model)
 
+SET(get_ready)
+
+SET(preconnect)
 SET(open)
 SET(close)
 
@@ -791,30 +844,21 @@ SET(can_write)
 SET(can_flush)
 SET(is_rotational)
 SET(can_trim)
-
-SET(dump_plugin)
-
 SET(can_zero)
 SET(can_fua)
+SET(can_multi_conn)
+SET(can_extents)
+SET(can_cache)
+SET(can_fast_zero)
 
 SET(pread)
 SET(pwrite)
 SET(flush)
 SET(trim)
 SET(zero)
-
-SET(can_multi_conn)
-
-SET(can_extents)
 SET(extents)
-
-SET(can_cache)
 SET(cache)
 
-SET(thread_model)
-
-SET(can_fast_zero)
-
 #undef SET
 
 static void
@@ -825,23 +869,30 @@ remove_roots (void)
   REMOVE (load);
   REMOVE (unload);
 
+  REMOVE (dump_plugin);
+
   REMOVE (config);
   REMOVE (config_complete);
+  REMOVE (thread_model);
 
+  REMOVE (get_ready);
+
+  REMOVE (preconnect);
   REMOVE (open);
   REMOVE (close);
 
   REMOVE (get_size);
 
-  REMOVE (can_write);
+  REMOVE (can_cache);
+  REMOVE (can_extents);
+  REMOVE (can_fast_zero);
   REMOVE (can_flush);
-  REMOVE (is_rotational);
-  REMOVE (can_trim);
-
-  REMOVE (dump_plugin);
-
-  REMOVE (can_zero);
   REMOVE (can_fua);
+  REMOVE (can_multi_conn);
+  REMOVE (can_trim);
+  REMOVE (can_write);
+  REMOVE (can_zero);
+  REMOVE (is_rotational);
 
   REMOVE (pread);
   REMOVE (pwrite);
@@ -849,18 +900,10 @@ remove_roots (void)
   REMOVE (trim);
   REMOVE (zero);
 
-  REMOVE (can_multi_conn);
-
-  REMOVE (can_extents);
   REMOVE (extents);
 
-  REMOVE (can_cache);
   REMOVE (cache);
 
-  REMOVE (thread_model);
-
-  REMOVE (can_fast_zero);
-
 #undef REMOVE
 }
 
@@ -941,6 +984,65 @@ ocaml_nbdkit_read_password (value strv)
   CAMLreturn (rv);
 }
 
+value
+ocaml_nbdkit_realpath (value strv)
+{
+  CAMLparam1 (strv);
+  CAMLlocal1 (rv);
+  char *ret;
+
+  ret = nbdkit_realpath (String_val (strv));
+  if (ret == NULL)
+    caml_failwith ("nbdkit_realpath");
+  rv = caml_copy_string (ret);
+  free (ret);
+
+  CAMLreturn (rv);
+}
+
+value
+ocaml_nbdkit_nanosleep (value secv, value nsecv)
+{
+  CAMLparam2 (secv, nsecv);
+  int r;
+
+  caml_enter_blocking_section ();
+  r = nbdkit_nanosleep (Int_val (secv), Int_val (nsecv));
+  caml_leave_blocking_section ();
+  if (r == -1)
+    caml_failwith ("nbdkit_nanosleep");
+
+  CAMLreturn (Val_unit);
+}
+
+value
+ocaml_nbdkit_export_name (value unitv)
+{
+  CAMLparam1 (unitv);
+  CAMLlocal1 (rv);
+  const char *ret;
+
+  ret = nbdkit_export_name ();
+  /* Note that NULL indicates error.  Default export name is [""] even
+   * for oldstyle.
+   */
+  if (ret == NULL)
+    caml_failwith ("nbdkit_export_name");
+  rv = caml_copy_string (ret);
+
+  CAMLreturn (rv);
+}
+
+/* NB: noalloc function. */
+value
+ocaml_nbdkit_shutdown (value unitv)
+{
+  CAMLparam1 (unitv);
+
+  nbdkit_shutdown ();
+  CAMLreturn (Val_unit);
+}
+
 /* NB: noalloc function. */
 value
 ocaml_nbdkit_debug (value strv)
