@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -46,6 +46,7 @@
 #define PY_SSIZE_T_CLEAN 1
 #include <Python.h>
 
+#define NBDKIT_API_VERSION 2
 #include <nbdkit-plugin.h>
 
 #include "cleanup.h"
@@ -60,8 +61,9 @@
  */
 static const char *script;
 static PyObject *module;
+static int py_api_version = 1;
 
-static int last_error;
+static __thread int last_error;
 
 static PyObject *
 set_error (PyObject *self, PyObject *args)
@@ -221,7 +223,7 @@ static struct PyModuleDef moduledef = {
   NULL
 };
 
-static PyMODINIT_FUNC
+PyMODINIT_FUNC
 create_nbdkit_module (void)
 {
   PyObject *m;
@@ -231,6 +233,36 @@ create_nbdkit_module (void)
     nbdkit_error ("could not create the nbdkit API module");
     exit (EXIT_FAILURE);
   }
+
+  /* Constants corresponding to various flags. */
+#define ADD_INT_CONSTANT(name)                                      \
+  if (PyModule_AddIntConstant (m, #name, NBDKIT_##name) == -1) {    \
+    nbdkit_error ("could not add constant %s to nbdkit API module", \
+                  #name);                                           \
+    exit (EXIT_FAILURE);                                            \
+  }
+  ADD_INT_CONSTANT (THREAD_MODEL_SERIALIZE_CONNECTIONS);
+  ADD_INT_CONSTANT (THREAD_MODEL_SERIALIZE_ALL_REQUESTS);
+  ADD_INT_CONSTANT (THREAD_MODEL_SERIALIZE_REQUESTS);
+  ADD_INT_CONSTANT (THREAD_MODEL_PARALLEL);
+
+  ADD_INT_CONSTANT (FLAG_MAY_TRIM);
+  ADD_INT_CONSTANT (FLAG_FUA);
+  ADD_INT_CONSTANT (FLAG_REQ_ONE);
+  ADD_INT_CONSTANT (FLAG_FAST_ZERO);
+
+  ADD_INT_CONSTANT (FUA_NONE);
+  ADD_INT_CONSTANT (FUA_EMULATE);
+  ADD_INT_CONSTANT (FUA_NATIVE);
+
+  ADD_INT_CONSTANT (CACHE_NONE);
+  ADD_INT_CONSTANT (CACHE_EMULATE);
+  ADD_INT_CONSTANT (CACHE_NATIVE);
+
+  ADD_INT_CONSTANT (EXTENT_HOLE);
+  ADD_INT_CONSTANT (EXTENT_ZERO);
+#undef ADD_INT_CONSTANT
+
   return m;
 }
 
@@ -245,7 +277,6 @@ static void
 py_unload (void)
 {
   Py_XDECREF (module);
-
   Py_Finalize ();
 }
 
@@ -255,9 +286,14 @@ py_dump_plugin (void)
   PyObject *fn;
   PyObject *r;
 
+  /* Python version and ABI. */
   printf ("python_version=%s\n", PY_VERSION);
   printf ("python_pep_384_abi_version=%d\n", PYTHON_ABI_VERSION);
 
+  /* Maximum nbdkit API version supported. */
+  printf ("nbdkit_python_maximum_api_version=%d\n", NBDKIT_API_VERSION);
+
+  /* If the script has a dump_plugin function, call it. */
   if (script && callback_defined ("dump_plugin", &fn)) {
     PyErr_Clear ();
 
@@ -267,6 +303,30 @@ py_dump_plugin (void)
   }
 }
 
+static int
+get_py_api_version (void)
+{
+  PyObject *obj;
+  long value;
+
+  obj = PyObject_GetAttrString (module, "API_VERSION");
+  if (obj == NULL)
+    return 1;                   /* Default to API version 1. */
+
+  value = PyLong_AsLong (obj);
+  Py_DECREF (obj);
+
+  if (value < 1 || value > NBDKIT_API_VERSION) {
+    nbdkit_error ("%s: API_VERSION requested unknown version: %ld.  "
+                  "This plugin supports API versions between 1 and %d.",
+                  script, value, NBDKIT_API_VERSION);
+    return -1;
+  }
+
+  nbdkit_debug ("module requested API_VERSION %ld", value);
+  return (int) value;
+}
+
 static int
 py_config (const char *key, const char *value)
 {
@@ -329,6 +389,11 @@ py_config (const char *key, const char *value)
                     "nbdkit requires these callbacks.", script);
       return -1;
     }
+
+    /* Get the API version. */
+    py_api_version = get_py_api_version ();
+    if (py_api_version == -1)
+      return -1;
   }
   else if (callback_defined ("config", &fn)) {
     /* Other parameters are passed to the Python .config callback. */
@@ -369,51 +434,86 @@ py_config_complete (void)
   return 0;
 }
 
+static int
+py_get_ready (void)
+{
+  PyObject *fn;
+  PyObject *r;
+
+  if (callback_defined ("get_ready", &fn)) {
+    PyErr_Clear ();
+
+    r = PyObject_CallObject (fn, NULL);
+    Py_DECREF (fn);
+    if (check_python_failure ("get_ready") == -1)
+      return -1;
+    Py_DECREF (r);
+  }
+
+  return 0;
+}
+
+struct handle {
+  int can_zero;
+  PyObject *py_h;
+};
+
 static void *
 py_open (int readonly)
 {
   PyObject *fn;
-  PyObject *handle;
+  struct handle *h;
 
   if (!callback_defined ("open", &fn)) {
     nbdkit_error ("%s: missing callback: %s", script, "open");
     return NULL;
   }
 
+  h = malloc (sizeof *h);
+  if (h == NULL) {
+    nbdkit_error ("malloc: %m");
+    return NULL;
+  }
+  h->can_zero = -1;
+
   PyErr_Clear ();
 
-  handle = PyObject_CallFunctionObjArgs (fn, readonly ? Py_True : Py_False,
-                                         NULL);
+  h->py_h = PyObject_CallFunctionObjArgs (fn, readonly ? Py_True : Py_False,
+                                          NULL);
   Py_DECREF (fn);
-  if (check_python_failure ("open") == -1)
+  if (check_python_failure ("open") == -1) {
+    free (h);
     return NULL;
+  }
 
-  return handle;
+  assert (h->py_h);
+  return h;
 }
 
 static void
 py_close (void *handle)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
 
   if (callback_defined ("close", &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
+    r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
     Py_DECREF (fn);
     check_python_failure ("close");
     Py_XDECREF (r);
   }
 
-  Py_DECREF (obj);
+  Py_DECREF (h->py_h);
+  free (h);
 }
 
 static int64_t
 py_get_size (void *handle)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
   int64_t ret;
@@ -425,7 +525,7 @@ py_get_size (void *handle)
 
   PyErr_Clear ();
 
-  r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
+  r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
   Py_DECREF (fn);
   if (check_python_failure ("get_size") == -1)
     return -1;
@@ -439,10 +539,10 @@ py_get_size (void *handle)
 }
 
 static int
-py_pread (void *handle, void *buf,
-          uint32_t count, uint64_t offset)
+py_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
+          uint32_t flags)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
   Py_buffer view = {0};
@@ -455,24 +555,40 @@ py_pread (void *handle, void *buf,
 
   PyErr_Clear ();
 
-  r = PyObject_CallFunction (fn, "OiL", obj, count, offset);
+  switch (py_api_version) {
+  case 1:
+    r = PyObject_CallFunction (fn, "OiL", h->py_h, count, offset);
+    break;
+  case 2:
+    r = PyObject_CallFunction (fn, "ONLI", h->py_h,
+          PyMemoryView_FromMemory ((char *)buf, count, PyBUF_WRITE),
+          offset, flags);
+    break;
+  default: abort ();
+  }
   Py_DECREF (fn);
   if (check_python_failure ("pread") == -1)
     return ret;
 
-  if (PyObject_GetBuffer (r, &view, PyBUF_SIMPLE) == -1) {
-    nbdkit_error ("%s: value returned from pread does not support the "
-                  "buffer protocol",
-                  script);
-    goto out;
-  }
+  if (py_api_version == 1) {
+    /* In API v1 the Python pread function had to return a buffer
+     * protocol compatible function.  In API v2+ it writes directly to
+     * the C buffer so this code is not used.
+     */
+    if (PyObject_GetBuffer (r, &view, PyBUF_SIMPLE) == -1) {
+      nbdkit_error ("%s: value returned from pread does not support the "
+                    "buffer protocol",
+                    script);
+      goto out;
+    }
 
-  if (view.len < count) {
-    nbdkit_error ("%s: buffer returned from pread is too small", script);
-    goto out;
-  }
+    if (view.len < count) {
+      nbdkit_error ("%s: buffer returned from pread is too small", script);
+      goto out;
+    }
 
-  memcpy (buf, view.buf, count);
+    memcpy (buf, view.buf, count);
+  }
   ret = 0;
 
 out:
@@ -485,19 +601,29 @@ out:
 }
 
 static int
-py_pwrite (void *handle, const void *buf,
-           uint32_t count, uint64_t offset)
+py_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
+           uint32_t flags)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
 
   if (callback_defined ("pwrite", &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunction (fn, "ONL", obj,
+    switch (py_api_version) {
+    case 1:
+      r = PyObject_CallFunction (fn, "ONL", h->py_h,
             PyMemoryView_FromMemory ((char *)buf, count, PyBUF_READ),
             offset);
+      break;
+    case 2:
+      r = PyObject_CallFunction (fn, "ONLI", h->py_h,
+            PyMemoryView_FromMemory ((char *)buf, count, PyBUF_READ),
+            offset, flags);
+      break;
+    default: abort ();
+    }
     Py_DECREF (fn);
     if (check_python_failure ("pwrite") == -1)
       return -1;
@@ -512,16 +638,24 @@ py_pwrite (void *handle, const void *buf,
 }
 
 static int
-py_flush (void *handle)
+py_flush (void *handle, uint32_t flags)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
 
   if (callback_defined ("flush", &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
+    switch (py_api_version) {
+    case 1:
+      r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
+      break;
+    case 2:
+      r = PyObject_CallFunction (fn, "OI", h->py_h, flags);
+      break;
+    default: abort ();
+    }
     Py_DECREF (fn);
     if (check_python_failure ("flush") == -1)
       return -1;
@@ -536,16 +670,24 @@ py_flush (void *handle)
 }
 
 static int
-py_trim (void *handle, uint32_t count, uint64_t offset)
+py_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
 
   if (callback_defined ("trim", &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunction (fn, "OiL", obj, count, offset);
+    switch (py_api_version) {
+    case 1:
+      r = PyObject_CallFunction (fn, "OiL", h->py_h, count, offset);
+      break;
+    case 2:
+      r = PyObject_CallFunction (fn, "OiLI", h->py_h, count, offset, flags);
+      break;
+    default: abort ();
+    }
     Py_DECREF (fn);
     if (check_python_failure ("trim") == -1)
       return -1;
@@ -560,9 +702,9 @@ py_trim (void *handle, uint32_t count, uint64_t offset)
 }
 
 static int
-py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+py_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
 
@@ -570,9 +712,19 @@ py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
     PyErr_Clear ();
 
     last_error = 0;
-    r = PyObject_CallFunction (fn, "OiLO",
-                               obj, count, offset,
-                               may_trim ? Py_True : Py_False);
+    switch (py_api_version) {
+    case 1: {
+      int may_trim = flags & NBDKIT_FLAG_MAY_TRIM;
+      r = PyObject_CallFunction (fn, "OiLO",
+                                 h->py_h, count, offset,
+                                 may_trim ? Py_True : Py_False);
+      break;
+    }
+    case 2:
+      r = PyObject_CallFunction (fn, "OiLI", h->py_h, count, offset, flags);
+      break;
+    default: abort ();
+    }
     Py_DECREF (fn);
     if (last_error == EOPNOTSUPP || last_error == ENOTSUP) {
       /* When user requests this particular error, we want to
@@ -595,10 +747,40 @@ py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
   return -1;
 }
 
+static int
+py_cache (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
+{
+  struct handle *h = handle;
+  PyObject *fn;
+  PyObject *r;
+
+  if (callback_defined ("cache", &fn)) {
+    PyErr_Clear ();
+
+    switch (py_api_version) {
+    case 1:
+    case 2:
+      r = PyObject_CallFunction (fn, "OiLI", h->py_h, count, offset, flags, NULL);
+      break;
+    default: abort ();
+    }
+    Py_DECREF (fn);
+    if (check_python_failure ("cache") == -1)
+      return -1;
+    Py_DECREF (r);
+  }
+  else {
+    nbdkit_error ("%s not implemented", "cache");
+    return -1;
+  }
+
+  return 0;
+}
+
 static int
 boolean_callback (void *handle, const char *can_fn, const char *plain_fn)
 {
-  PyObject *obj = handle;
+  struct handle *h = handle;
   PyObject *fn;
   PyObject *r;
   int ret;
@@ -606,7 +788,7 @@ boolean_callback (void *handle, const char *can_fn, const char *plain_fn)
   if (callback_defined (can_fn, &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
+    r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
     Py_DECREF (fn);
     if (check_python_failure (can_fn) == -1)
       return -1;
@@ -630,6 +812,12 @@ py_is_rotational (void *handle)
   return boolean_callback (handle, "is_rotational", NULL);
 }
 
+static int
+py_can_multi_conn (void *handle)
+{
+  return boolean_callback (handle, "can_multi_conn", NULL);
+}
+
 static int
 py_can_write (void *handle)
 {
@@ -648,6 +836,89 @@ py_can_trim (void *handle)
   return boolean_callback (handle, "can_trim", "trim");
 }
 
+static int
+py_can_zero (void *handle)
+{
+  struct handle *h = handle;
+
+  if (h->can_zero >= 0)
+    return h->can_zero;
+  return h->can_zero = boolean_callback (handle, "can_zero", "zero");
+}
+
+static int
+py_can_fast_zero (void *handle)
+{
+  int r;
+
+  if (callback_defined ("can_fast_zero", NULL))
+    return boolean_callback (handle, "can_fast_zero", NULL);
+
+  /* No Python ‘can_fast_zero’, but we advertise fast fail support when
+   * 'can_zero' is false.  (In C modules, nbdkit would do this).
+   */
+  r = py_can_zero (handle);
+  if (r == -1)
+    return -1;
+  return !r;
+}
+
+static int
+py_can_fua (void *handle)
+{
+  struct handle *h = handle;
+  PyObject *fn;
+  PyObject *r;
+  int ret;
+
+  if (callback_defined ("can_fua", &fn)) {
+    PyErr_Clear ();
+
+    r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
+    Py_DECREF (fn);
+    if (check_python_failure ("can_fua") == -1)
+      return -1;
+    ret = PyLong_AsLong (r);
+    Py_DECREF (r);
+    return ret;
+  }
+  /* No Python ‘can_fua’, but check if there's a Python ‘flush’
+   * callback defined.  (In C modules, nbdkit would do this).
+   */
+  else if (callback_defined ("flush", NULL))
+    return NBDKIT_FUA_EMULATE;
+  else
+    return NBDKIT_FUA_NONE;
+}
+
+static int
+py_can_cache (void *handle)
+{
+  struct handle *h = handle;
+  PyObject *fn;
+  PyObject *r;
+  int ret;
+
+  if (callback_defined ("can_cache", &fn)) {
+    PyErr_Clear ();
+
+    r = PyObject_CallFunctionObjArgs (fn, h->py_h, NULL);
+    Py_DECREF (fn);
+    if (check_python_failure ("can_cache") == -1)
+      return -1;
+    ret = PyLong_AsLong (r);
+    Py_DECREF (r);
+    return ret;
+  }
+  /* No Python ‘can_cache’, but check if there's a Python ‘cache’
+   * callback defined.  (In C modules, nbdkit would do this).
+   */
+  else if (callback_defined ("cache", NULL))
+    return NBDKIT_CACHE_NATIVE;
+  else
+    return NBDKIT_CACHE_NONE;
+}
+
 #define py_config_help \
   "script=<FILENAME>     (required) The Python plugin to run.\n" \
   "[other arguments may be used by the plugin that you load]"
@@ -666,20 +937,28 @@ static struct nbdkit_plugin plugin = {
   .config_complete   = py_config_complete,
   .config_help       = py_config_help,
 
+  .get_ready         = py_get_ready,
+
   .open              = py_open,
   .close             = py_close,
 
   .get_size          = py_get_size,
   .is_rotational     = py_is_rotational,
+  .can_multi_conn    = py_can_multi_conn,
   .can_write         = py_can_write,
   .can_flush         = py_can_flush,
   .can_trim          = py_can_trim,
+  .can_zero          = py_can_zero,
+  .can_fast_zero     = py_can_fast_zero,
+  .can_fua           = py_can_fua,
+  .can_cache         = py_can_cache,
 
   .pread             = py_pread,
   .pwrite            = py_pwrite,
   .flush             = py_flush,
   .trim              = py_trim,
   .zero              = py_zero,
+  .cache             = py_cache,
 };
 
 NBDKIT_REGISTER_PLUGIN (plugin)
