@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -39,7 +39,9 @@
 #include <inttypes.h>
 #include <string.h>
 #include <unistd.h>
+#include <assert.h>
 #include <dlfcn.h>
+#include <libgen.h>
 
 #define NBDKIT_API_VERSION 2
 
@@ -50,16 +52,16 @@
 #include "minmax.h"
 #include "rounding.h"
 
+#include "vddk.h"
 #include "vddk-structs.h"
 
-/* Enable extra disk info debugging with: -D vddk.diskinfo=1 */
+/* Debug flags. */
 int vddk_debug_diskinfo;
-
-/* Enable debugging of extents code with: -D vddk.extents=1 */
 int vddk_debug_extents;
+int vddk_debug_datapath = 1;
 
 /* For each VDDK API define a static global variable.  These globals
- * are initialized when the plugin is loaded (by vddk_load).
+ * are initialized when the plugin is loaded (by vddk_get_ready).
  */
 #define STUB(fn,ret,args) static ret (*fn) args
 #define OPTIONAL_STUB(fn,ret,args) static ret (*fn) args
@@ -69,27 +71,27 @@ int vddk_debug_extents;
 
 /* Parameters passed to InitEx. */
 #define VDDK_MAJOR 5
-#define VDDK_MINOR 1
+#define VDDK_MINOR 5
 
-static void *dl = NULL;                    /* dlopen handle */
-static int init_called = 0;                /* was InitEx called */
+static void *dl;                           /* dlopen handle */
+static bool init_called;                   /* was InitEx called */
 
-static char *config = NULL;                /* config */
-static const char *cookie = NULL;          /* cookie */
-static const char *filename = NULL;        /* file */
-static char *libdir = NULL;                /* libdir */
-static uint16_t nfc_host_port = 0;         /* nfchostport */
-static char *password = NULL;              /* password */
-static uint16_t port = 0;                  /* port */
-static const char *server_name = NULL;     /* server */
-static bool single_link = false;           /* single-link */
-static const char *snapshot_moref = NULL;  /* snapshot */
-static const char *thumb_print = NULL;     /* thumbprint */
-static const char *transport_modes = NULL; /* transports */
-static bool unbuffered = false;            /* unbuffered */
-static const char *username = NULL;        /* user */
-static const char *vmx_spec = NULL;        /* vm */
-static bool is_remote = false;
+static char *config;                       /* config */
+static const char *cookie;                 /* cookie */
+static const char *filename;               /* file */
+char *libdir;                              /* libdir */
+static uint16_t nfc_host_port;             /* nfchostport */
+char *password;                            /* password */
+static uint16_t port;                      /* port */
+static const char *server_name;            /* server */
+static bool single_link;                   /* single-link */
+static const char *snapshot_moref;         /* snapshot */
+static const char *thumb_print;            /* thumbprint */
+static const char *transport_modes;        /* transports */
+static bool unbuffered;                    /* unbuffered */
+static const char *username;               /* user */
+static const char *vmx_spec;               /* vm */
+static bool is_remote;
 
 #define VDDK_ERROR(err, fs, ...)                                \
   do {                                                          \
@@ -101,6 +103,9 @@ static bool is_remote = false;
 
 #define DEBUG_CALL(fn, fs, ...)                                 \
   nbdkit_debug ("VDDK call: %s (" fs ")", fn, ##__VA_ARGS__)
+#define DEBUG_CALL_DATAPATH(fn, fs, ...)                        \
+  if (vddk_debug_datapath)                                      \
+    nbdkit_debug ("VDDK call: %s (" fs ")", fn, ##__VA_ARGS__)
 
 static void
 trim (char *str)
@@ -140,57 +145,19 @@ error_function (const char *fs, va_list args)
 
   trim (str);
 
+  /* VDDK 7 added a useless error message about their "phone home"
+   * system called CEIP which only panics users.  Demote it to a debug
+   * statement.  https://bugzilla.redhat.com/show_bug.cgi?id=1834267
+   */
+  if (strstr (str, "Get CEIP status failed") != NULL) {
+    nbdkit_debug ("%s", str);
+    return;
+  }
+
   nbdkit_error ("%s", str);
 }
 
-/* Load and unload the plugin. */
-static void
-vddk_load (void)
-{
-  static const char *sonames[] = {
-    /* Prefer the newest library in case multiple exist. */
-    "libvixDiskLib.so.6",
-    "libvixDiskLib.so.5",
-  };
-  size_t i;
-  CLEANUP_FREE char *orig_error = NULL;
-
-  /* Load the library. */
-  for (i = 0; i < sizeof sonames / sizeof sonames[0]; ++i) {
-    dl = dlopen (sonames[i], RTLD_NOW);
-    if (dl != NULL)
-      break;
-    if (i == 0) {
-      orig_error = dlerror ();
-      if (orig_error)
-        orig_error = strdup (orig_error);
-    }
-  }
-  if (dl == NULL) {
-    nbdkit_error ("%s\n\n"
-                  "If '%s' is located on a non-standard path you may need to\n"
-                  "set $LD_LIBRARY_PATH or edit /etc/ld.so.conf.\n\n"
-                  "See the nbdkit-vddk-plugin(1) man page for details.",
-                  orig_error ? : "(unknown error)", sonames[0]);
-    exit (EXIT_FAILURE);
-  }
-
-  /* Load symbols. */
-#define STUB(fn,ret,args)                                         \
-  do {                                                            \
-    fn = dlsym (dl, #fn);                                         \
-    if (fn == NULL) {                                             \
-      nbdkit_error ("required VDDK symbol \"%s\" is missing: %s", \
-                    #fn, dlerror ());                             \
-      exit (EXIT_FAILURE);                                        \
-    }                                                             \
-  } while (0)
-#define OPTIONAL_STUB(fn,ret,args) fn = dlsym (dl, #fn)
-#include "vddk-stubs.h"
-#undef STUB
-#undef OPTIONAL_STUB
-}
-
+/* Unload the plugin. */
 static void
 vddk_unload (void)
 {
@@ -209,6 +176,8 @@ vddk_unload (void)
 static int
 vddk_config (const char *key, const char *value)
 {
+  int r;
+
   if (strcmp (key, "config") == 0) {
     /* See FILENAMES AND PATHS in nbdkit-plugin(3). */
     free (config);
@@ -237,6 +206,17 @@ vddk_config (const char *key, const char *value)
     if (nbdkit_parse_uint16_t ("nfchostport", value, &nfc_host_port) == -1)
       return -1;
   }
+  else if (strcmp (key, "noreexec") == 0) {
+    /* This undocumented option disables reexec.  The caller must set
+     * LD_LIBRARY_PATH correctly as for older versions of the plugin.
+     * This option is only for use when debugging reexec, eg. to see
+     * if it causing a problem.
+     */
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    noreexec = r;
+  }
   else if (strcmp (key, "password") == 0) {
     free (password);
     if (nbdkit_read_password (value, &password) == -1)
@@ -246,12 +226,15 @@ vddk_config (const char *key, const char *value)
     if (nbdkit_parse_uint16_t ("port", value, &port) == -1)
       return -1;
   }
+  else if (strcmp (key, "reexeced_") == 0) {
+    /* Special name because it is only for internal use. */
+    reexeced = (char *)value;
+  }
   else if (strcmp (key, "server") == 0) {
     server_name = value;
   }
   else if (strcmp (key, "single-link") == 0) {
-    int r = nbdkit_parse_bool (value);
-
+    r = nbdkit_parse_bool (value);
     if (r == -1)
       return -1;
     single_link = r;
@@ -266,8 +249,7 @@ vddk_config (const char *key, const char *value)
     transport_modes = value;
   }
   else if (strcmp (key, "unbuffered") == 0) {
-    int r = nbdkit_parse_bool (value);
-
+    r = nbdkit_parse_bool (value);
     if (r == -1)
       return -1;
     unbuffered = r;
@@ -289,11 +271,92 @@ vddk_config (const char *key, const char *value)
   return 0;
 }
 
+/* Load the VDDK library. */
+static void
+load_library (bool load_error_is_fatal)
+{
+  static const char *sonames[] = {
+    /* Prefer the newest library in case multiple exist.  Check two
+     * possible directories: the usual VDDK installation puts .so
+     * files in an arch-specific subdirectory of $libdir (our minimum
+     * supported version is VDDK 5.5.5, which only supports x64-64);
+     * but our testsuite is easier to write if we point libdir
+     * directly to a stub .so.
+     */
+    "lib64/libvixDiskLib.so.7",
+    "libvixDiskLib.so.7",
+    "lib64/libvixDiskLib.so.6",
+    "libvixDiskLib.so.6",
+    "lib64/libvixDiskLib.so.5",
+    "libvixDiskLib.so.5",
+  };
+  size_t i;
+  CLEANUP_FREE char *orig_error = NULL;
+
+  if (!libdir) {
+    libdir = strdup (VDDK_LIBDIR);
+    if (!libdir) {
+      nbdkit_error ("strdup: %m");
+      exit (EXIT_FAILURE);
+    }
+  }
+
+  for (i = 0; i < sizeof sonames / sizeof sonames[0]; ++i) {
+    CLEANUP_FREE char *path;
+
+    /* Set the full path so that dlopen will preferentially load the
+     * system libraries from the same directory.
+     */
+    if (asprintf (&path, "%s/%s", libdir, sonames[i]) == -1) {
+      nbdkit_error ("asprintf: %m");
+      exit (EXIT_FAILURE);
+    }
+
+    dl = dlopen (path, RTLD_NOW);
+    if (dl != NULL) {
+      /* Now that we found the library, ensure that LD_LIBRARY_PATH
+       * includes its directory for all future loads.  This may modify
+       * path in-place and/or re-exec nbdkit, but that's okay.
+       */
+      reexec_if_needed (dirname (path));
+      break;
+    }
+    if (i == 0) {
+      orig_error = dlerror ();
+      if (orig_error)
+        orig_error = strdup (orig_error);
+    }
+  }
+  if (dl == NULL) {
+    if (!load_error_is_fatal)
+      return;
+    nbdkit_error ("%s\n\n"
+                  "If '%s' is located on a non-standard path you may need to\n"
+                  "set libdir=/path/to/vmware-vix-disklib-distrib.\n\n"
+                  "See the nbdkit-vddk-plugin(1) man page for details.",
+                  orig_error ? : "(unknown error)", sonames[0]);
+    exit (EXIT_FAILURE);
+  }
+
+  /* Load symbols. */
+#define STUB(fn,ret,args)                                         \
+  do {                                                            \
+    fn = dlsym (dl, #fn);                                         \
+    if (fn == NULL) {                                             \
+      nbdkit_error ("required VDDK symbol \"%s\" is missing: %s", \
+                    #fn, dlerror ());                             \
+      exit (EXIT_FAILURE);                                        \
+    }                                                             \
+  } while (0)
+#define OPTIONAL_STUB(fn,ret,args) fn = dlsym (dl, #fn)
+#include "vddk-stubs.h"
+#undef STUB
+#undef OPTIONAL_STUB
+}
+
 static int
 vddk_config_complete (void)
 {
-  VixError err;
-
   if (filename == NULL) {
     nbdkit_error ("you must supply the file=<FILENAME> parameter "
                   "after the plugin name on the command line");
@@ -330,32 +393,48 @@ vddk_config_complete (void)
 #undef missing
   }
 
+  /* Restore original LD_LIBRARY_PATH after reexec. */
+  if (restore_ld_library_path () == -1)
+    return -1;
+
+  return 0;
+}
+
+static int
+vddk_get_ready (void)
+{
+  VixError err;
+
+  load_library (true);
+
   /* Initialize VDDK library. */
   DEBUG_CALL ("VixDiskLib_InitEx",
               "%d, %d, &debug_fn, &error_fn, &error_fn, %s, %s",
               VDDK_MAJOR, VDDK_MINOR,
-              libdir ? : VDDK_LIBDIR, config ? : "NULL");
+              libdir, config ? : "NULL");
   err = VixDiskLib_InitEx (VDDK_MAJOR, VDDK_MINOR,
                            &debug_function, /* log function */
                            &error_function, /* warn function */
                            &error_function, /* panic function */
-                           libdir ? : VDDK_LIBDIR, config);
+                           libdir, config);
   if (err != VIX_OK) {
     VDDK_ERROR (err, "VixDiskLib_InitEx");
     exit (EXIT_FAILURE);
   }
-  init_called = 1;
+  init_called = true;
 
   return 0;
 }
 
 #define vddk_config_help \
-  "file=<FILENAME>     (required) The filename (eg. VMDK file) to serve.\n" \
+  "[file=]<FILENAME>   (required) The filename (eg. VMDK file) to serve.\n" \
   "Many optional parameters are supported, see nbdkit-vddk-plugin(3)."
 
 static void
 vddk_dump_plugin (void)
 {
+  load_library (false);
+
   printf ("vddk_default_libdir=%s\n", VDDK_LIBDIR);
   printf ("vddk_has_nfchostport=1\n");
 
@@ -368,7 +447,8 @@ vddk_dump_plugin (void)
    */
   Dl_info info;
   CLEANUP_FREE char *p = NULL;
-  if (dladdr (VixDiskLib_InitEx, &info) != 0 &&
+  if (dl != NULL &&
+      dladdr (VixDiskLib_InitEx, &info) != 0 &&
       info.dli_fname != NULL &&
       (p = nbdkit_realpath (info.dli_fname)) != NULL) {
     printf ("vddk_dll=%s\n", p);
@@ -455,6 +535,7 @@ vddk_open (int readonly)
     h->params->thumbPrint = (char *) thumb_print;
     h->params->port = port;
     h->params->nfcHostPort = nfc_host_port;
+    h->params->specType = VIXDISKLIB_SPEC_VMX;
   }
 
   /* XXX Some documentation suggests we should call
@@ -580,19 +661,20 @@ vddk_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
 
   /* Align to sectors. */
   if (!IS_ALIGNED (offset, VIXDISKLIB_SECTOR_SIZE)) {
-    nbdkit_error ("read is not aligned to sectors");
+    nbdkit_error ("%s is not aligned to sectors", "read");
     return -1;
   }
   if (!IS_ALIGNED (count, VIXDISKLIB_SECTOR_SIZE)) {
-    nbdkit_error ("read is not aligned to sectors");
+    nbdkit_error ("%s is not aligned to sectors", "read");
     return -1;
   }
   offset /= VIXDISKLIB_SECTOR_SIZE;
   count /= VIXDISKLIB_SECTOR_SIZE;
 
-  DEBUG_CALL ("VixDiskLib_Read",
-              "handle, %" PRIu64 " sectors, %" PRIu32 " sectors, buffer",
-              offset, count);
+  DEBUG_CALL_DATAPATH ("VixDiskLib_Read",
+                       "handle, %" PRIu64 " sectors, "
+                       "%" PRIu32 " sectors, buffer",
+                       offset, count);
   err = VixDiskLib_Read (h->handle, offset, count, buf);
   if (err != VIX_OK) {
     VDDK_ERROR (err, "VixDiskLib_Read");
@@ -618,19 +700,20 @@ vddk_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
 
   /* Align to sectors. */
   if (!IS_ALIGNED (offset, VIXDISKLIB_SECTOR_SIZE)) {
-    nbdkit_error ("read is not aligned to sectors");
+    nbdkit_error ("%s is not aligned to sectors", "write");
     return -1;
   }
   if (!IS_ALIGNED (count, VIXDISKLIB_SECTOR_SIZE)) {
-    nbdkit_error ("read is not aligned to sectors");
+    nbdkit_error ("%s is not aligned to sectors", "write");
     return -1;
   }
   offset /= VIXDISKLIB_SECTOR_SIZE;
   count /= VIXDISKLIB_SECTOR_SIZE;
 
-  DEBUG_CALL ("VixDiskLib_Write",
-              "handle, %" PRIu64 " sectors, %" PRIu32 " sectors, buffer",
-              offset, count);
+  DEBUG_CALL_DATAPATH ("VixDiskLib_Write",
+                       "handle, %" PRIu64 " sectors, "
+                       "%" PRIu32 " sectors, buffer",
+                       offset, count);
   err = VixDiskLib_Write (h->handle, offset, count, buf);
   if (err != VIX_OK) {
     VDDK_ERROR (err, "VixDiskLib_Write");
@@ -831,12 +914,13 @@ static struct nbdkit_plugin plugin = {
   .name              = "vddk",
   .longname          = "VMware VDDK plugin",
   .version           = PACKAGE_VERSION,
-  .load              = vddk_load,
   .unload            = vddk_unload,
   .config            = vddk_config,
   .config_complete   = vddk_config_complete,
   .config_help       = vddk_config_help,
+  .magic_config_key  = "file",
   .dump_plugin       = vddk_dump_plugin,
+  .get_ready         = vddk_get_ready,
   .open              = vddk_open,
   .close             = vddk_close,
   .get_size          = vddk_get_size,
