@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2019 Red Hat Inc.
+ * Copyright (C) 2019-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -151,7 +151,13 @@ run_command (void)
   }
 
   if (pid > 0) {              /* Parent process is the run command. */
-    r = system (cmd);
+    /* Restore original stdin/out */
+    if (dup2 (saved_stdin, STDIN_FILENO) == -1 ||
+        dup2 (saved_stdout, STDOUT_FILENO) == -1) {
+      r = -1;
+    }
+    else
+      r = system (cmd);
     if (r == -1) {
       nbdkit_error ("failure to execute external command: %m");
       r = EXIT_FAILURE;
@@ -171,12 +177,14 @@ run_command (void)
       r = EXIT_FAILURE;
       break;
     case 0:
-      /* Captive nbdkit still running; kill it, but no need to wait
-       * for it, as the --run program's exit status is good enough (if
-       * the captive nbdkit fails to exit after SIGTERM, we have a
-       * bigger bug to fix).
+      /* Captive nbdkit is still running; kill it.  We want to wait
+       * for nbdkit to exit since that ensures all cleanup is done in
+       * the plugin before we return.  However we don't care if nbdkit
+       * returns an error, the exit code we return always comes from
+       * the --run command.
        */
       kill (pid, SIGTERM);
+      waitpid (pid, NULL, 0);
       break;
     default:
       /* Captive nbdkit exited unexpectedly; update the exit status. */
