@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -53,23 +53,15 @@ static struct connection *new_connection (int sockin, int sockout,
 static void free_connection (struct connection *conn);
 
 /* Don't call these raw socket functions directly.  Use conn->recv etc. */
-static int raw_recv (struct connection *, void *buf, size_t len);
-static int raw_send_socket (struct connection *, const void *buf, size_t len,
-                            int flags);
-static int raw_send_other (struct connection *, const void *buf, size_t len,
-                           int flags);
-static void raw_close (struct connection *);
-
-void *
-connection_get_handle (struct connection *conn, size_t i)
-{
-  assert (i < conn->nr_handles);
-  return conn->handles[i].handle;
-}
+static int raw_recv ( void *buf, size_t len);
+static int raw_send_socket (const void *buf, size_t len, int flags);
+static int raw_send_other (const void *buf, size_t len, int flags);
+static void raw_close (void);
 
 int
-connection_get_status (struct connection *conn)
+connection_get_status (void)
 {
+  GET_CONN;
   int r;
 
   if (conn->nworkers &&
@@ -86,8 +78,10 @@ connection_get_status (struct connection *conn)
  * For convenience, return the incoming value.
  */
 int
-connection_set_status (struct connection *conn, int value)
+connection_set_status (int value)
 {
+  GET_CONN;
+
   if (conn->nworkers &&
       pthread_mutex_lock (&conn->status_lock))
     abort ();
@@ -125,8 +119,8 @@ connection_worker (void *data)
   threadlocal_set_conn (conn);
   free (worker);
 
-  while (!quit && connection_get_status (conn) > 0)
-    protocol_recv_request_send_reply (conn);
+  while (!quit && connection_get_status () > 0)
+    protocol_recv_request_send_reply ();
   debug ("exiting worker thread %s", threadlocal_get_name ());
   free (name);
   return NULL;
@@ -143,35 +137,42 @@ handle_single_connection (int sockin, int sockout)
 
   lock_connection ();
 
-  if (backend->thread_model (backend) < NBDKIT_THREAD_MODEL_PARALLEL ||
-      nworkers == 1)
+  /* Because of asynchronous exit it is plausible that a new
+   * connection is started at the same time as the backend is being
+   * shut down.  top may therefore be NULL, and if this happens return
+   * immediately.
+   */
+  if (!top) {
+    unlock_connection ();
+    return;
+  }
+
+  if (thread_model < NBDKIT_THREAD_MODEL_PARALLEL || nworkers == 1)
     nworkers = 0;
   conn = new_connection (sockin, sockout, nworkers);
   if (!conn)
     goto done;
 
-  /* NB: because of an asynchronous exit backend can be set to NULL at
-   * just about any time.
-   */
-  if (backend)
-    plugin_name = backend->plugin_name (backend);
-  else
-    plugin_name = "(unknown)";
+  plugin_name = top->plugin_name (top);
   threadlocal_set_name (plugin_name);
 
+  if (top->preconnect (top, read_only) == -1)
+    goto done;
+
   /* NBD handshake.
    *
    * Note that this calls the backend .open callback when it is safe
    * to do so (eg. after TLS authentication).
    */
-  if (protocol_handshake (conn) == -1)
+  if (protocol_handshake () == -1)
     goto done;
+  conn->handshake_complete = true;
 
   if (!nworkers) {
     /* No need for a separate thread. */
     debug ("handshake complete, processing requests serially");
-    while (!quit && connection_get_status (conn) > 0)
-      protocol_recv_request_send_reply (conn);
+    while (!quit && connection_get_status () > 0)
+      protocol_recv_request_send_reply ();
   }
   else {
     /* Create thread pool to process requests. */
@@ -189,13 +190,13 @@ handle_single_connection (int sockin, int sockout)
 
       if (unlikely (!worker)) {
         perror ("malloc");
-        connection_set_status (conn, -1);
+        connection_set_status (-1);
         goto wait;
       }
       if (unlikely (asprintf (&worker->name, "%s.%d", plugin_name, nworkers)
                     < 0)) {
         perror ("asprintf");
-        connection_set_status (conn, -1);
+        connection_set_status (-1);
         free (worker);
         goto wait;
       }
@@ -205,7 +206,7 @@ handle_single_connection (int sockin, int sockout)
       if (unlikely (err)) {
         errno = err;
         perror ("pthread_create");
-        connection_set_status (conn, -1);
+        connection_set_status (-1);
         free (worker);
         goto wait;
       }
@@ -218,9 +219,9 @@ handle_single_connection (int sockin, int sockout)
   }
 
   /* Finalize (for filters), called just before close. */
-  lock_request (conn);
-  r = backend_finalize (backend, conn);
-  unlock_request (conn);
+  lock_request ();
+  r = backend_finalize (top);
+  unlock_request ();
   if (r == -1)
     goto done;
 
@@ -242,17 +243,21 @@ new_connection (int sockin, int sockout, int nworkers)
     perror ("malloc");
     return NULL;
   }
-
   conn->status_pipe[0] = conn->status_pipe[1] = -1;
 
-  conn->handles = calloc (backend->i + 1, sizeof *conn->handles);
+  pthread_mutex_init (&conn->request_lock, NULL);
+  pthread_mutex_init (&conn->read_lock, NULL);
+  pthread_mutex_init (&conn->write_lock, NULL);
+  pthread_mutex_init (&conn->status_lock, NULL);
+
+  conn->handles = calloc (top->i + 1, sizeof *conn->handles);
   if (conn->handles == NULL) {
     perror ("malloc");
-    goto error;
+    goto error1;
   }
-  conn->nr_handles = backend->i + 1;
+  conn->nr_handles = top->i + 1;
   for_each_backend (b)
-    reset_b_conn_handle (&conn->handles[b->i]);
+    reset_handle (get_handle (conn, b->i));
 
   conn->status = 1;
   conn->nworkers = nworkers;
@@ -260,7 +265,7 @@ new_connection (int sockin, int sockout, int nworkers)
 #ifdef HAVE_PIPE2
     if (pipe2 (conn->status_pipe, O_NONBLOCK | O_CLOEXEC)) {
       perror ("pipe2");
-      goto error;
+      goto error2;
     }
 #else
     /* If we were fully parallel, then this function could be
@@ -271,36 +276,31 @@ new_connection (int sockin, int sockout, int nworkers)
      * we aren't accepting until the plugin is not running, making
      * non-atomicity okay.
      */
-    assert (backend->thread_model (backend) <=
-            NBDKIT_THREAD_MODEL_SERIALIZE_ALL_REQUESTS);
-    lock_request (NULL);
+    assert (thread_model <= NBDKIT_THREAD_MODEL_SERIALIZE_ALL_REQUESTS);
+    lock_request ();
     if (pipe (conn->status_pipe)) {
       perror ("pipe");
-      unlock_request (NULL);
-      goto error;
+      unlock_request ();
+      goto error2;
     }
     if (set_nonblock (set_cloexec (conn->status_pipe[0])) == -1) {
       perror ("fcntl");
       close (conn->status_pipe[1]);
-      unlock_request (NULL);
-      goto error;
+      unlock_request ();
+      goto error2;
     }
     if (set_nonblock (set_cloexec (conn->status_pipe[1])) == -1) {
       perror ("fcntl");
       close (conn->status_pipe[0]);
-      unlock_request (NULL);
-      goto error;
+      unlock_request ();
+      goto error2;
     }
-    unlock_request (NULL);
+    unlock_request ();
 #endif
   }
+
   conn->sockin = sockin;
   conn->sockout = sockout;
-  pthread_mutex_init (&conn->request_lock, NULL);
-  pthread_mutex_init (&conn->read_lock, NULL);
-  pthread_mutex_init (&conn->write_lock, NULL);
-  pthread_mutex_init (&conn->status_lock, NULL);
-
   conn->recv = raw_recv;
   if (getsockopt (sockout, SOL_SOCKET, SO_TYPE, &opt, &optlen) == 0)
     conn->send = raw_send_socket;
@@ -312,12 +312,18 @@ new_connection (int sockin, int sockout, int nworkers)
 
   return conn;
 
- error:
+ error2:
   if (conn->status_pipe[0] >= 0)
     close (conn->status_pipe[0]);
   if (conn->status_pipe[1] >= 0)
     close (conn->status_pipe[1]);
   free (conn->handles);
+
+ error1:
+  pthread_mutex_destroy (&conn->request_lock);
+  pthread_mutex_destroy (&conn->read_lock);
+  pthread_mutex_destroy (&conn->write_lock);
+  pthread_mutex_destroy (&conn->status_lock);
   free (conn);
   return NULL;
 }
@@ -328,29 +334,16 @@ free_connection (struct connection *conn)
   if (!conn)
     return;
 
-  threadlocal_set_conn (NULL);
-  conn->close (conn);
-  if (listen_stdin) {
-    int fd;
-
-    /* Restore something to stdin/out so the rest of our code can
-     * continue to assume that all new fds will be above stderr.
-     * Swap directions to get EBADF on improper use of stdin/out.
-     */
-    fd = open ("/dev/null", O_WRONLY | O_CLOEXEC);
-    assert (fd == 0);
-    fd = open ("/dev/null", O_RDONLY | O_CLOEXEC);
-    assert (fd == 1);
-  }
+  conn->close ();
 
   /* Don't call the plugin again if quit has been set because the main
    * thread will be in the process of unloading it.  The plugin.unload
    * callback should always be called.
    */
   if (!quit) {
-    lock_request (conn);
-    backend_close (backend, conn);
-    unlock_request (conn);
+    lock_request ();
+    backend_close (top);
+    unlock_request ();
   }
 
   if (conn->status_pipe[0] >= 0) {
@@ -365,6 +358,7 @@ free_connection (struct connection *conn)
 
   free (conn->handles);
   free (conn);
+  threadlocal_set_conn (NULL);
 }
 
 /* Write buffer to conn->sockout with send() and either succeed completely
@@ -372,9 +366,9 @@ free_connection (struct connection *conn)
  * that this send will be followed by related data.
  */
 static int
-raw_send_socket (struct connection *conn, const void *vbuf, size_t len,
-                 int flags)
+raw_send_socket (const void *vbuf, size_t len, int flags)
 {
+  GET_CONN;
   int sock = conn->sockout;
   const char *buf = vbuf;
   ssize_t r;
@@ -402,9 +396,9 @@ raw_send_socket (struct connection *conn, const void *vbuf, size_t len,
  * (returns 0) or fail (returns -1). flags is ignored.
  */
 static int
-raw_send_other (struct connection *conn, const void *vbuf, size_t len,
-                int flags)
+raw_send_other (const void *vbuf, size_t len, int flags)
 {
+  GET_CONN;
   int sock = conn->sockout;
   const char *buf = vbuf;
   ssize_t r;
@@ -427,8 +421,9 @@ raw_send_other (struct connection *conn, const void *vbuf, size_t len,
  * (returns > 0), read an EOF (returns 0), or fail (returns -1).
  */
 static int
-raw_recv (struct connection *conn, void *vbuf, size_t len)
+raw_recv (void *vbuf, size_t len)
 {
+  GET_CONN;
   int sock = conn->sockin;
   char *buf = vbuf;
   ssize_t r;
@@ -460,8 +455,10 @@ raw_recv (struct connection *conn, void *vbuf, size_t len)
  * close, so this function ignores errors.
  */
 static void
-raw_close (struct connection *conn)
+raw_close (void)
 {
+  GET_CONN;
+
   if (conn->sockin >= 0)
     close (conn->sockin);
   if (conn->sockout >= 0 && conn->sockin != conn->sockout)
