@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2019 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -49,14 +49,6 @@ struct backend_filter {
   struct nbdkit_filter filter;
 };
 
-/* Literally a backend + a connection pointer.  This is the
- * implementation of ‘void *nxdata’ in the filter API.
- */
-struct b_conn {
-  struct backend *b;
-  struct connection *conn;
-};
-
 /* Note this frees the whole chain. */
 static void
 filter_free (struct backend *b)
@@ -74,7 +66,7 @@ filter_thread_model (struct backend *b)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
   int filter_thread_model = NBDKIT_THREAD_MODEL_PARALLEL;
-  int thread_model = b->next->thread_model (b->next);
+  int model = b->next->thread_model (b->next);
 
   if (f->filter.thread_model) {
     filter_thread_model = f->filter.thread_model ();
@@ -82,10 +74,10 @@ filter_thread_model (struct backend *b)
       exit (EXIT_FAILURE);
   }
 
-  if (filter_thread_model < thread_model) /* more serialized */
-    thread_model = filter_thread_model;
+  if (filter_thread_model < model) /* more serialized */
+    model = filter_thread_model;
 
-  return thread_model;
+  return model;
 }
 
 /* This is actually passing the request through to the final plugin,
@@ -135,9 +127,8 @@ filter_dump_fields (struct backend *b)
 }
 
 static int
-next_config (void *nxdata, const char *key, const char *value)
+next_config (struct backend *b, const char *key, const char *value)
 {
-  struct backend *b = nxdata;
   b->config (b, key, value);
   return 0;
 }
@@ -159,9 +150,8 @@ filter_config (struct backend *b, const char *key, const char *value)
 }
 
 static int
-next_config_complete (void *nxdata)
+next_config_complete (struct backend *b)
 {
-  struct backend *b = nxdata;
   b->config_complete (b);
   return 0;
 }
@@ -181,6 +171,41 @@ filter_config_complete (struct backend *b)
     b->next->config_complete (b->next);
 }
 
+static int
+next_get_ready (struct backend *b)
+{
+  b->get_ready (b);
+  return 0;
+}
+
+static void
+filter_get_ready (struct backend *b)
+{
+  struct backend_filter *f = container_of (b, struct backend_filter, backend);
+
+  debug ("%s: get_ready", b->name);
+
+  if (f->filter.get_ready) {
+    if (f->filter.get_ready (next_get_ready, b->next) == -1)
+      exit (EXIT_FAILURE);
+  }
+  else
+    b->next->get_ready (b->next);
+}
+
+static int
+filter_preconnect (struct backend *b, int readonly)
+{
+  struct backend_filter *f = container_of (b, struct backend_filter, backend);
+
+  debug ("%s: preconnect", b->name);
+
+  if (f->filter.preconnect)
+    return f->filter.preconnect (b->next->preconnect, b->next, readonly);
+  else
+    return b->next->preconnect (b->next, readonly);
+}
+
 /* magic_config_key only applies to plugins, so this passes the
  * request through to the plugin (hence the name).
  */
@@ -190,33 +215,26 @@ plugin_magic_config_key (struct backend *b)
   return b->next->magic_config_key (b->next);
 }
 
-static int
-next_open (void *nxdata, int readonly)
-{
-  struct b_conn *b_conn = nxdata;
-
-  return backend_open (b_conn->b, b_conn->conn, readonly);
-}
-
 static void *
-filter_open (struct backend *b, struct connection *conn, int readonly)
+filter_open (struct backend *b, int readonly)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
+  void *handle;
 
   /* Most filters will call next_open first, resulting in
    * inner-to-outer ordering.
    */
   if (f->filter.open)
-    return f->filter.open (next_open, &nxdata, readonly);
-  else if (backend_open (b->next, conn, readonly) == -1)
-    return NULL;
+    handle = f->filter.open (backend_open, b->next, readonly);
+  else if (backend_open (b->next, readonly) == -1)
+    handle = NULL;
   else
-    return NBDKIT_HANDLE_NOT_NEEDED;
+    handle = NBDKIT_HANDLE_NOT_NEEDED;
+  return handle;
 }
 
 static void
-filter_close (struct backend *b, struct connection *conn, void *handle)
+filter_close (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
 
@@ -224,438 +242,267 @@ filter_close (struct backend *b, struct connection *conn, void *handle)
     f->filter.close (handle);
 }
 
-/* The next_functions structure contains pointers to backend
- * functions.  However because these functions are all expecting a
- * backend and a connection, we cannot call them directly, but must
- * write some next_* functions that unpack the two parameters from a
- * single ‘void *nxdata’ struct pointer (‘b_conn’).
- */
-
-static int
-next_reopen (void *nxdata, int readonly)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_reopen (b_conn->b, b_conn->conn, readonly);
-}
-
-static int64_t
-next_get_size (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_get_size (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_write (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_write (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_flush (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_flush (b_conn->b, b_conn->conn);
-}
-
-static int
-next_is_rotational (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_is_rotational (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_trim (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_trim (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_zero (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_zero (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_fast_zero (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_fast_zero (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_extents (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_extents (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_fua (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_fua (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_multi_conn (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_multi_conn (b_conn->b, b_conn->conn);
-}
-
-static int
-next_can_cache (void *nxdata)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_can_cache (b_conn->b, b_conn->conn);
-}
-
-static int
-next_pread (void *nxdata, void *buf, uint32_t count, uint64_t offset,
-            uint32_t flags, int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_pread (b_conn->b, b_conn->conn, buf, count, offset, flags,
-                        err);
-}
-
-static int
-next_pwrite (void *nxdata, const void *buf, uint32_t count, uint64_t offset,
-             uint32_t flags, int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_pwrite (b_conn->b, b_conn->conn, buf, count, offset, flags,
-                         err);
-}
-
-static int
-next_flush (void *nxdata, uint32_t flags, int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_flush (b_conn->b, b_conn->conn, flags, err);
-}
-
-static int
-next_trim (void *nxdata, uint32_t count, uint64_t offset, uint32_t flags,
-           int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_trim (b_conn->b, b_conn->conn, count, offset, flags, err);
-}
-
-static int
-next_zero (void *nxdata, uint32_t count, uint64_t offset, uint32_t flags,
-           int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_zero (b_conn->b, b_conn->conn, count, offset, flags, err);
-}
-
-static int
-next_extents (void *nxdata, uint32_t count, uint64_t offset, uint32_t flags,
-              struct nbdkit_extents *extents, int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_extents (b_conn->b, b_conn->conn, count, offset, flags,
-                          extents, err);
-}
-
-static int
-next_cache (void *nxdata, uint32_t count, uint64_t offset,
-            uint32_t flags, int *err)
-{
-  struct b_conn *b_conn = nxdata;
-  return backend_cache (b_conn->b, b_conn->conn, count, offset, flags, err);
-}
-
 static struct nbdkit_next_ops next_ops = {
-  .reopen = next_reopen,
-  .get_size = next_get_size,
-  .can_write = next_can_write,
-  .can_flush = next_can_flush,
-  .is_rotational = next_is_rotational,
-  .can_trim = next_can_trim,
-  .can_zero = next_can_zero,
-  .can_fast_zero = next_can_fast_zero,
-  .can_extents = next_can_extents,
-  .can_fua = next_can_fua,
-  .can_multi_conn = next_can_multi_conn,
-  .can_cache = next_can_cache,
-  .pread = next_pread,
-  .pwrite = next_pwrite,
-  .flush = next_flush,
-  .trim = next_trim,
-  .zero = next_zero,
-  .extents = next_extents,
-  .cache = next_cache,
+  .reopen = backend_reopen,
+  .get_size = backend_get_size,
+  .can_write = backend_can_write,
+  .can_flush = backend_can_flush,
+  .is_rotational = backend_is_rotational,
+  .can_trim = backend_can_trim,
+  .can_zero = backend_can_zero,
+  .can_fast_zero = backend_can_fast_zero,
+  .can_extents = backend_can_extents,
+  .can_fua = backend_can_fua,
+  .can_multi_conn = backend_can_multi_conn,
+  .can_cache = backend_can_cache,
+  .pread = backend_pread,
+  .pwrite = backend_pwrite,
+  .flush = backend_flush,
+  .trim = backend_trim,
+  .zero = backend_zero,
+  .extents = backend_extents,
+  .cache = backend_cache,
 };
 
 static int
-filter_prepare (struct backend *b, struct connection *conn, void *handle,
-                int readonly)
+filter_prepare (struct backend *b, void *handle, int readonly)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.prepare &&
-      f->filter.prepare (&next_ops, &nxdata, handle, readonly) == -1)
+      f->filter.prepare (&next_ops, b->next, handle, readonly) == -1)
     return -1;
 
   return 0;
 }
 
 static int
-filter_finalize (struct backend *b, struct connection *conn, void *handle)
+filter_finalize (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.finalize &&
-      f->filter.finalize (&next_ops, &nxdata, handle) == -1)
+      f->filter.finalize (&next_ops, b->next, handle) == -1)
     return -1;
   return 0;
 }
 
 static int64_t
-filter_get_size (struct backend *b, struct connection *conn, void *handle)
+filter_get_size (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.get_size)
-    return f->filter.get_size (&next_ops, &nxdata, handle);
+    return f->filter.get_size (&next_ops, b->next, handle);
   else
-    return backend_get_size (b->next, conn);
+    return backend_get_size (b->next);
 }
 
 static int
-filter_can_write (struct backend *b, struct connection *conn, void *handle)
+filter_can_write (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_write)
-    return f->filter.can_write (&next_ops, &nxdata, handle);
+    return f->filter.can_write (&next_ops, b->next, handle);
   else
-    return backend_can_write (b->next, conn);
+    return backend_can_write (b->next);
 }
 
 static int
-filter_can_flush (struct backend *b, struct connection *conn, void *handle)
+filter_can_flush (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_flush)
-    return f->filter.can_flush (&next_ops, &nxdata, handle);
+    return f->filter.can_flush (&next_ops, b->next, handle);
   else
-    return backend_can_flush (b->next, conn);
+    return backend_can_flush (b->next);
 }
 
 static int
-filter_is_rotational (struct backend *b, struct connection *conn, void *handle)
+filter_is_rotational (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.is_rotational)
-    return f->filter.is_rotational (&next_ops, &nxdata, handle);
+    return f->filter.is_rotational (&next_ops, b->next, handle);
   else
-    return backend_is_rotational (b->next, conn);
+    return backend_is_rotational (b->next);
 }
 
 static int
-filter_can_trim (struct backend *b, struct connection *conn, void *handle)
+filter_can_trim (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_trim)
-    return f->filter.can_trim (&next_ops, &nxdata, handle);
+    return f->filter.can_trim (&next_ops, b->next, handle);
   else
-    return backend_can_trim (b->next, conn);
+    return backend_can_trim (b->next);
 }
 
 static int
-filter_can_zero (struct backend *b, struct connection *conn, void *handle)
+filter_can_zero (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_zero)
-    return f->filter.can_zero (&next_ops, &nxdata, handle);
+    return f->filter.can_zero (&next_ops, b->next, handle);
   else
-    return backend_can_zero (b->next, conn);
+    return backend_can_zero (b->next);
 }
 
 static int
-filter_can_fast_zero (struct backend *b, struct connection *conn, void *handle)
+filter_can_fast_zero (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_fast_zero)
-    return f->filter.can_fast_zero (&next_ops, &nxdata, handle);
+    return f->filter.can_fast_zero (&next_ops, b->next, handle);
   else
-    return backend_can_fast_zero (b->next, conn);
+    return backend_can_fast_zero (b->next);
 }
 
 static int
-filter_can_extents (struct backend *b, struct connection *conn, void *handle)
+filter_can_extents (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_extents)
-    return f->filter.can_extents (&next_ops, &nxdata, handle);
+    return f->filter.can_extents (&next_ops, b->next, handle);
   else
-    return backend_can_extents (b->next, conn);
+    return backend_can_extents (b->next);
 }
 
 static int
-filter_can_fua (struct backend *b, struct connection *conn, void *handle)
+filter_can_fua (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_fua)
-    return f->filter.can_fua (&next_ops, &nxdata, handle);
+    return f->filter.can_fua (&next_ops, b->next, handle);
   else
-    return backend_can_fua (b->next, conn);
+    return backend_can_fua (b->next);
 }
 
 static int
-filter_can_multi_conn (struct backend *b, struct connection *conn, void *handle)
+filter_can_multi_conn (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_multi_conn)
-    return f->filter.can_multi_conn (&next_ops, &nxdata, handle);
+    return f->filter.can_multi_conn (&next_ops, b->next, handle);
   else
-    return backend_can_multi_conn (b->next, conn);
+    return backend_can_multi_conn (b->next);
 }
 
 static int
-filter_can_cache (struct backend *b, struct connection *conn, void *handle)
+filter_can_cache (struct backend *b, void *handle)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.can_cache)
-    return f->filter.can_cache (&next_ops, &nxdata, handle);
+    return f->filter.can_cache (&next_ops, b->next, handle);
   else
-    return backend_can_cache (b->next, conn);
+    return backend_can_cache (b->next);
 }
 
 static int
-filter_pread (struct backend *b, struct connection *conn, void *handle,
+filter_pread (struct backend *b, void *handle,
               void *buf, uint32_t count, uint64_t offset,
               uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.pread)
-    return f->filter.pread (&next_ops, &nxdata, handle,
+    return f->filter.pread (&next_ops, b->next, handle,
                             buf, count, offset, flags, err);
   else
-    return backend_pread (b->next, conn, buf, count, offset, flags, err);
+    return backend_pread (b->next, buf, count, offset, flags, err);
 }
 
 static int
-filter_pwrite (struct backend *b, struct connection *conn, void *handle,
+filter_pwrite (struct backend *b, void *handle,
                const void *buf, uint32_t count, uint64_t offset,
                uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.pwrite)
-    return f->filter.pwrite (&next_ops, &nxdata, handle,
+    return f->filter.pwrite (&next_ops, b->next, handle,
                              buf, count, offset, flags, err);
   else
-    return backend_pwrite (b->next, conn, buf, count, offset, flags, err);
+    return backend_pwrite (b->next, buf, count, offset, flags, err);
 }
 
 static int
-filter_flush (struct backend *b, struct connection *conn, void *handle,
+filter_flush (struct backend *b, void *handle,
               uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.flush)
-    return f->filter.flush (&next_ops, &nxdata, handle, flags, err);
+    return f->filter.flush (&next_ops, b->next, handle, flags, err);
   else
-    return backend_flush (b->next, conn, flags, err);
+    return backend_flush (b->next, flags, err);
 }
 
 static int
-filter_trim (struct backend *b, struct connection *conn, void *handle,
+filter_trim (struct backend *b, void *handle,
              uint32_t count, uint64_t offset,
              uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.trim)
-    return f->filter.trim (&next_ops, &nxdata, handle, count, offset, flags,
-                           err);
+    return f->filter.trim (&next_ops, b->next, handle, count, offset,
+                           flags, err);
   else
-    return backend_trim (b->next, conn, count, offset, flags, err);
+    return backend_trim (b->next, count, offset, flags, err);
 }
 
 static int
-filter_zero (struct backend *b, struct connection *conn, void *handle,
+filter_zero (struct backend *b, void *handle,
              uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.zero)
-    return f->filter.zero (&next_ops, &nxdata, handle,
+    return f->filter.zero (&next_ops, b->next, handle,
                            count, offset, flags, err);
   else
-    return backend_zero (b->next, conn, count, offset, flags, err);
+    return backend_zero (b->next, count, offset, flags, err);
 }
 
 static int
-filter_extents (struct backend *b, struct connection *conn, void *handle,
+filter_extents (struct backend *b, void *handle,
                 uint32_t count, uint64_t offset, uint32_t flags,
                 struct nbdkit_extents *extents, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
 
   if (f->filter.extents)
-    return f->filter.extents (&next_ops, &nxdata, handle,
+    return f->filter.extents (&next_ops, b->next, handle,
                               count, offset, flags,
                               extents, err);
   else
-    return backend_extents (b->next, conn, count, offset, flags,
+    return backend_extents (b->next, count, offset, flags,
                             extents, err);
 }
 
 static int
-filter_cache (struct backend *b, struct connection *conn, void *handle,
+filter_cache (struct backend *b, void *handle,
               uint32_t count, uint64_t offset,
               uint32_t flags, int *err)
 {
   struct backend_filter *f = container_of (b, struct backend_filter, backend);
-  struct b_conn nxdata = { .b = b->next, .conn = conn };
-
 
   if (f->filter.cache)
-    return f->filter.cache (&next_ops, &nxdata, handle,
+    return f->filter.cache (&next_ops, b->next, handle,
                             count, offset, flags, err);
   else
-    return backend_cache (b->next, conn, count, offset, flags, err);
+    return backend_cache (b->next, count, offset, flags, err);
 }
 
 static struct backend filter_functions = {
@@ -668,6 +515,8 @@ static struct backend filter_functions = {
   .config = filter_config,
   .config_complete = filter_config_complete,
   .magic_config_key = plugin_magic_config_key,
+  .get_ready = filter_get_ready,
+  .preconnect = filter_preconnect,
   .open = filter_open,
   .prepare = filter_prepare,
   .finalize = filter_finalize,
