@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -45,15 +45,31 @@
 #include <syslog.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/socket.h>
+
+#ifdef HAVE_SYS_MMAN_H
+#include <sys/mman.h>
+#endif
+
+#ifdef HAVE_LINUX_VM_SOCKETS_H
+#include <linux/vm_sockets.h>
+#endif
 
 #include <pthread.h>
 
 #include <dlfcn.h>
 
+#include "ascii-string.h"
+
 #include "internal.h"
+#include "nbd-protocol.h"
 #include "options.h"
 #include "exit-with-parent.h"
 
+#ifdef ENABLE_LIBFUZZER
+#define main fuzzer_main
+#endif
+
 static char *make_random_fifo (void);
 static struct backend *open_plugin_so (size_t i, const char *filename, int short_name);
 static struct backend *open_filter_so (struct backend *next, size_t i, const char *filename, int short_name);
@@ -67,14 +83,17 @@ const char *exportname;         /* -e */
 bool foreground;                /* -f */
 const char *ipaddr;             /* -i */
 enum log_to log_to = LOG_TO_DEFAULT; /* --log */
+unsigned mask_handshake = ~0U;  /* --mask-handshake */
 bool newstyle = true;           /* false = -o, true = -n */
+bool no_sr;                     /* --no-sr */
 char *pidfile;                  /* -P */
 const char *port;               /* -p */
-bool readonly;                  /* -r */
+bool read_only;                 /* -r */
 const char *run;                /* --run */
 bool listen_stdin;              /* -s */
 const char *selinux_label;      /* --selinux-label */
-int threads;                    /* -t */
+bool swap;                      /* --swap */
+unsigned threads;               /* -t */
 int tls;                        /* --tls : 0=off 1=on 2=require */
 const char *tls_certificates_dir; /* --tls-certificates */
 const char *tls_psk;            /* --tls-psk */
@@ -82,10 +101,14 @@ bool tls_verify_peer;           /* --tls-verify-peer */
 char *unixsocket;               /* -U */
 const char *user, *group;       /* -u & -g */
 bool verbose;                   /* -v */
-unsigned int socket_activation  /* $LISTEN_FDS and $LISTEN_PID set */;
+bool vsock;                     /* --vsock */
+unsigned int socket_activation; /* $LISTEN_FDS and $LISTEN_PID set */
+bool configured;                /* .config_complete done */
+int saved_stdin = -1;           /* dup'd stdin during -s/--run */
+int saved_stdout = -1;          /* dup'd stdout during -s/--run */
 
-/* The currently loaded plugin. */
-struct backend *backend;
+/* The linked list of zero or more filters, and one plugin. */
+struct backend *top;
 
 static char *random_fifo_dir = NULL;
 static char *random_fifo = NULL;
@@ -94,11 +117,11 @@ static void
 usage (void)
 {
   /* --{short,long}-options remain undocumented */
-  printf (
+  const char *opt_list =
 #include "synopsis.c"
-  );
-  printf ("\n"
-          "Please read the nbdkit(1) manual page for full usage.\n");
+	  ;
+  printf ("%s\n", opt_list);
+  printf ("Please read the nbdkit(1) manual page for full usage.\n");
 }
 
 static void
@@ -110,6 +133,10 @@ display_version (void)
 static void
 dump_config (void)
 {
+  CLEANUP_FREE char *binary = realpath ("/proc/self/exe", NULL);
+
+  if (binary != NULL)
+    printf ("%s=%s\n", "binary", binary);
   printf ("%s=%s\n", "bindir", bindir);
   printf ("%s=%s\n", "filterdir", filterdir);
   printf ("%s=%s\n", "libdir", libdir);
@@ -130,6 +157,8 @@ dump_config (void)
   printf ("tls=no\n");
 #endif
   printf ("%s=%s\n", "version", PACKAGE_VERSION);
+  printf ("%s=%d\n", "version_major", NBDKIT_VERSION_MAJOR);
+  printf ("%s=%d\n", "version_minor", NBDKIT_VERSION_MINOR);
 }
 
 int
@@ -138,7 +167,7 @@ main (int argc, char *argv[])
   int c;
   bool help = false, version = false, dump_plugin = false;
   int tls_set_on_cli = false;
-  int short_name;
+  bool short_name;
   const char *filename;
   char *p;
   static struct filter_filename {
@@ -148,7 +177,26 @@ main (int argc, char *argv[])
   size_t i;
   const char *magic_config_key;
 
+  /* Refuse to run if stdin/out/err are closed, whether or not -s is used. */
+  if (fcntl (STDERR_FILENO, F_GETFL) == -1) {
+    /* Nowhere we can report the error. Oh well. */
+    exit (EXIT_FAILURE);
+  }
+  if (fcntl (STDIN_FILENO, F_GETFL) == -1 ||
+      fcntl (STDOUT_FILENO, F_GETFL) == -1) {
+    perror ("expecting stdin/stdout to be opened");
+    exit (EXIT_FAILURE);
+  }
+
+#if !ENABLE_LIBFUZZER
   threadlocal_init ();
+#else
+  static bool main_called = false;
+  if (!main_called) {
+    threadlocal_init ();
+    main_called = true;
+  }
+#endif
 
   /* The default setting for TLS depends on whether we were
    * compiled with GnuTLS.
@@ -168,51 +216,6 @@ main (int argc, char *argv[])
       break;
 
     switch (c) {
-    case 'D':
-      {
-        const char *p, *q;
-        struct debug_flag *flag;
-
-        /* Debug Flag must be "NAME.FLAG=N".
-         *                     ^    ^    ^
-         *                optarg    p    q  (after +1 adjustment below)
-         */
-        p = strchr (optarg, '.');
-        q = strchr (optarg, '=');
-        if (p == NULL || q == NULL) {
-        bad_debug_flag:
-          fprintf (stderr,
-                   "%s: -D (Debug Flag) must have the format NAME.FLAG=N\n",
-                   program_name);
-          exit (EXIT_FAILURE);
-        }
-        p++;                    /* +1 adjustment */
-        q++;
-
-        if (p - optarg <= 1) goto bad_debug_flag; /* NAME too short */
-        if (p > q) goto bad_debug_flag;
-        if (q - p <= 1) goto bad_debug_flag; /* FLAG too short */
-        if (*q == '\0') goto bad_debug_flag; /* N too short */
-
-        flag = malloc (sizeof *flag);
-        if (flag == NULL) {
-        debug_flag_perror:
-          perror ("malloc");
-          exit (EXIT_FAILURE);
-        }
-        flag->name = strndup (optarg, p-optarg-1);
-        if (!flag->name) goto debug_flag_perror;
-        flag->flag = strndup (p, q-p-1);
-        if (!flag->flag) goto debug_flag_perror;
-        if (sscanf (q, "%d", &flag->value) != 1) goto bad_debug_flag;
-        flag->used = false;
-
-        /* Add flag to the linked list. */
-        flag->next = debug_flags;
-        debug_flags = flag;
-      }
-      break;
-
     case DUMP_CONFIG_OPTION:
       dump_config ();
       exit (EXIT_SUCCESS);
@@ -254,8 +257,11 @@ main (int argc, char *argv[])
         log_to = LOG_TO_STDERR;
       else if (strcmp (optarg, "syslog") == 0)
         log_to = LOG_TO_SYSLOG;
+      else if (strcmp (optarg, "null") == 0)
+        log_to = LOG_TO_NULL;
       else {
-        fprintf (stderr, "%s: --log must be \"stderr\" or \"syslog\"\n",
+        fprintf (stderr, "%s: "
+                 "--log must be \"stderr\", \"syslog\" or \"null\"\n",
                  program_name);
         exit (EXIT_FAILURE);
       }
@@ -290,11 +296,15 @@ main (int argc, char *argv[])
       }
       exit (EXIT_SUCCESS);
 
+    case SWAP_OPTION:
+      swap = 1;
+      break;
+
     case TLS_OPTION:
       tls_set_on_cli = true;
-      if (strcasecmp (optarg, "require") == 0 ||
-          strcasecmp (optarg, "required") == 0 ||
-          strcasecmp (optarg, "force") == 0)
+      if (ascii_strcasecmp (optarg, "require") == 0 ||
+          ascii_strcasecmp (optarg, "required") == 0 ||
+          ascii_strcasecmp (optarg, "force") == 0)
         tls = 2;
       else {
         tls = nbdkit_parse_bool (optarg);
@@ -315,8 +325,27 @@ main (int argc, char *argv[])
       tls_verify_peer = true;
       break;
 
+    case VSOCK_OPTION:
+#ifdef AF_VSOCK
+      vsock = true;
+      break;
+#else
+      fprintf (stderr, "%s: AF_VSOCK is not supported on this platform\n",
+               program_name);
+      exit (EXIT_FAILURE);
+#endif
+
+    case 'D':
+      add_debug_flag (optarg);
+      break;
+
     case 'e':
       exportname = optarg;
+      if (strnlen (exportname, NBD_MAX_STRING + 1) > NBD_MAX_STRING) {
+        nbdkit_error ("export name too long");
+        exit (EXIT_FAILURE);
+      }
+      /* TODO: Check that name is valid UTF-8? */
       newstyle = true;
       break;
 
@@ -337,10 +366,20 @@ main (int argc, char *argv[])
       ipaddr = optarg;
       break;
 
+    case MASK_HANDSHAKE_OPTION:
+      if (nbdkit_parse_unsigned ("mask-handshake",
+                                 optarg, &mask_handshake) == -1)
+        exit (EXIT_FAILURE);
+      break;
+
     case 'n':
       newstyle = true;
       break;
 
+    case NO_SR_OPTION:
+      no_sr = true;
+      break;
+
     case 'o':
       newstyle = false;
       break;
@@ -361,7 +400,7 @@ main (int argc, char *argv[])
       break;
 
     case 'r':
-      readonly = true;
+      read_only = true;
       break;
 
     case 's':
@@ -374,18 +413,9 @@ main (int argc, char *argv[])
       break;
 
     case 't':
-      {
-        char *end;
-
-        errno = 0;
-        threads = strtoul (optarg, &end, 0);
-        if (errno || *end) {
-          fprintf (stderr, "%s: cannot parse '%s' into threads\n",
-                   program_name, optarg);
-          exit (EXIT_FAILURE);
-        }
-        /* XXX Worth a maximimum limit on threads? */
-      }
+      if (nbdkit_parse_unsigned ("threads", optarg, &threads) == -1)
+        exit (EXIT_FAILURE);
+      /* XXX Worth a maximimum limit on threads? */
       break;
 
     case 'U':
@@ -498,6 +528,27 @@ main (int argc, char *argv[])
   }
 #endif
 
+  /* If the user has mixed up -p/--run/-s/-U/--vsock options, then
+   * give an error.
+   *
+   * XXX Actually the server could easily be extended to handle both
+   * TCP/IP and Unix sockets, or even multiple TCP/IP ports.
+   */
+  if ((port && unixsocket) ||
+      (port && listen_stdin) ||
+      (unixsocket && listen_stdin) ||
+      (listen_stdin && run) ||
+      (listen_stdin && dump_plugin) ||
+      (vsock && unixsocket) ||
+      (vsock && listen_stdin) ||
+      (vsock && run)) {
+    fprintf (stderr,
+             "%s: --dump-plugin, -p, --run, -s, -U or --vsock options "
+             "cannot be used in this combination\n",
+             program_name);
+    exit (EXIT_FAILURE);
+  }
+
   /* The remaining command line arguments are the plugin name and
    * parameters.  If --help, --version or --dump-plugin were specified
    * then we open the plugin so that we can display the per-plugin
@@ -510,7 +561,6 @@ main (int argc, char *argv[])
    * If so we simply execute it with the current command line.
    */
   if (short_name) {
-    size_t i;
     struct stat statbuf;
     CLEANUP_FREE char *script;
 
@@ -537,39 +587,28 @@ main (int argc, char *argv[])
 
   /* Open the plugin (first) and then wrap the plugin with the
    * filters.  The filters are wrapped in reverse order that they
-   * appear on the command line so that in the end ‘backend’ points to
+   * appear on the command line so that in the end ‘top’ points to
    * the first filter on the command line.
    */
-  backend = open_plugin_so (0, filename, short_name);
+  top = open_plugin_so (0, filename, short_name);
   i = 1;
   while (filter_filenames) {
     struct filter_filename *t = filter_filenames;
-    const char *filename = t->filename;
-    int short_name = is_short_name (filename);
 
-    backend = open_filter_so (backend, i++, filename, short_name);
+    filename = t->filename;
+    short_name = is_short_name (filename);
+
+    top = open_filter_so (top, i++, filename, short_name);
 
     filter_filenames = t->next;
     free (t);
   }
 
+  /* Apply nbdkit.* flags for the server. */
+  apply_debug_flags (NULL, "nbdkit");
+
   /* Check all debug flags were used, and free them. */
-  while (debug_flags != NULL) {
-    struct debug_flag *next = debug_flags->next;
-
-    if (!debug_flags->used) {
-      fprintf (stderr, "%s: debug flag -D %s.%s was not used\n",
-               program_name, debug_flags->name, debug_flags->flag);
-      exit (EXIT_FAILURE);
-    }
-    free (debug_flags->name);
-    free (debug_flags->flag);
-    free (debug_flags);
-    debug_flags = next;
-  }
-
-  /* Select a thread model. */
-  lock_init_thread_model ();
+  free_debug_flags ();
 
   if (help) {
     struct backend *b;
@@ -579,7 +618,7 @@ main (int argc, char *argv[])
       printf ("\n");
       b->usage (b);
     }
-    backend->free (backend);
+    top->free (top);
     exit (EXIT_SUCCESS);
   }
 
@@ -589,12 +628,12 @@ main (int argc, char *argv[])
 
     display_version ();
     for_each_backend (b) {
-      printf ("%s", b->name (b));
+      printf ("%s", b->name);
       if ((v = b->version (b)) != NULL)
         printf (" %s", v);
       printf ("\n");
     }
-    backend->free (backend);
+    top->free (top);
     exit (EXIT_SUCCESS);
   }
 
@@ -607,17 +646,35 @@ main (int argc, char *argv[])
    * scripting languages, if magic_config_key == NULL then if the
    * first parameter is bare it is prefixed with the key "script", and
    * any other bare parameters are errors.
+   *
+   * Keys must live for the life of nbdkit.  Since we want to avoid
+   * modifying argv (so that /proc/PID/cmdline remains sane) but we
+   * need to create a key from argv[i] = "key=value" we must save the
+   * keys in an array which is freed at the end of main().
    */
-  magic_config_key = backend->magic_config_key (backend);
+  char **keys = calloc (argc, sizeof (char *));
+  if (keys == NULL) {
+    perror ("calloc");
+    exit (EXIT_FAILURE);
+  }
+
+  magic_config_key = top->magic_config_key (top);
   for (i = 0; optind < argc; ++i, ++optind) {
+    size_t n;
+
     p = strchr (argv[optind], '=');
-    if (p && is_config_key (argv[optind], p - argv[optind])) { /* key=value */
-      *p = '\0';
-      backend->config (backend, argv[optind], p+1);
+    n = p - argv[optind];
+    if (p && is_config_key (argv[optind], n)) { /* Is it key=value? */
+      keys[optind] = strndup (argv[optind], n);
+      if (keys[optind] == NULL) {
+        perror ("strndup");
+        exit (EXIT_FAILURE);
+      }
+      top->config (top, keys[optind], p+1);
     }
     else if (magic_config_key == NULL) {
       if (i == 0)               /* magic script parameter */
-        backend->config (backend, "script", argv[optind]);
+        top->config (top, "script", argv[optind]);
       else {
         fprintf (stderr,
                  "%s: expecting key=value on the command line but got: %s\n",
@@ -626,7 +683,7 @@ main (int argc, char *argv[])
       }
     }
     else {                      /* magic config key */
-      backend->config (backend, magic_config_key, argv[optind]);
+      top->config (top, magic_config_key, argv[optind]);
     }
   }
 
@@ -636,17 +693,63 @@ main (int argc, char *argv[])
    * parameters.
    */
   if (dump_plugin) {
-    backend->dump_fields (backend);
-    backend->free (backend);
+    top->dump_fields (top);
+    top->free (top);
+    for (i = 1; i < argc; ++i)
+      free (keys[i]);
+    free (keys);
     exit (EXIT_SUCCESS);
   }
 
-  backend->config_complete (backend);
+  top->config_complete (top);
+
+  /* Select the correct thread model based on config. */
+  lock_init_thread_model ();
+
+  /* Sanitize stdin/stdout to /dev/null, after saving the originals
+   * when needed.  We are still single-threaded at this point, and
+   * already checked that stdin/out were open, so we don't have to
+   * worry about other threads accidentally grabbing our intended fds,
+   * or races on FD_CLOEXEC.  POSIX says that 'fflush(NULL)' is
+   * supposed to reset the underlying offset of seekable stdin, but
+   * glibc is buggy and requires an explicit fflush(stdin) as
+   * well. https://sourceware.org/bugzilla/show_bug.cgi?id=12799
+   */
+  fflush (stdin);
+  fflush (NULL);
+  if (listen_stdin || run) {
+#ifndef F_DUPFD_CLOEXEC
+#define F_DUPFD_CLOEXEC F_DUPFD
+#endif
+    saved_stdin = fcntl (STDIN_FILENO, F_DUPFD_CLOEXEC, STDERR_FILENO + 1);
+    saved_stdout = fcntl (STDOUT_FILENO, F_DUPFD_CLOEXEC, STDERR_FILENO + 1);
+#if F_DUPFD == F_DUPFD_CLOEXEC
+    saved_stdin = set_cloexec (saved_stdin);
+    saved_stdout = set_cloexec (saved_stdout);
+#endif
+    if (saved_stdin == -1 || saved_stdout == -1) {
+      perror ("fcntl");
+      exit (EXIT_FAILURE);
+    }
+  }
+  close (STDIN_FILENO);
+  close (STDOUT_FILENO);
+  if (open ("/dev/null", O_RDONLY) != STDIN_FILENO ||
+      open ("/dev/null", O_WRONLY) != STDOUT_FILENO) {
+    perror ("open");
+    exit (EXIT_FAILURE);
+  }
+
+  /* Tell the plugin that we are about to start serving.  This must be
+   * called before we change user, fork, or open any sockets.
+   */
+  configured = true;
+  top->get_ready (top);
 
   start_serving ();
 
-  backend->free (backend);
-  backend = NULL;
+  top->free (top);
+  top = NULL;
 
   free (unixsocket);
   free (pidfile);
@@ -662,8 +765,16 @@ main (int argc, char *argv[])
   }
 
   crypto_free ();
+  close_quit_pipe ();
 
-  exit (EXIT_SUCCESS);
+  for (i = 1; i < argc; ++i)
+    free (keys[i]);
+  free (keys);
+
+  /* Note: Don't exit here, otherwise this won't work when compiled
+   * for libFuzzer.
+   */
+  return EXIT_SUCCESS;
 }
 
 /* Implementation of '-U -' */
@@ -671,7 +782,7 @@ static char *
 make_random_fifo (void)
 {
   char template[] = "/tmp/nbdkitXXXXXX";
-  char *unixsocket;
+  char *sock;
 
   if (mkdtemp (template) == NULL) {
     perror ("mkdtemp");
@@ -689,13 +800,13 @@ make_random_fifo (void)
     return NULL;
   }
 
-  unixsocket = strdup (random_fifo);
-  if (unixsocket == NULL) {
+  sock = strdup (random_fifo);
+  if (sock == NULL) {
     perror ("strdup");
     return NULL;
   }
 
-  return unixsocket;
+  return sock;
 }
 
 static struct backend *
@@ -802,44 +913,48 @@ open_filter_so (struct backend *next, size_t i,
 static void
 start_serving (void)
 {
-  int *socks;
-  size_t nr_socks;
+  sockets socks = empty_vector;
   size_t i;
 
-  /* If the user has mixed up -p/-U/-s options, then give an error.
-   *
-   * XXX Actually the server could easily be extended to handle both
-   * TCP/IP and Unix sockets, or even multiple TCP/IP ports.
-   */
-  if ((port && unixsocket) || (port && listen_stdin) ||
-      (unixsocket && listen_stdin) || (listen_stdin && run)) {
-    fprintf (stderr,
-             "%s: -p, -U and -s options cannot appear at the same time\n",
-             program_name);
-    exit (EXIT_FAILURE);
-  }
-
   set_up_quit_pipe ();
+#if !ENABLE_LIBFUZZER
   set_up_signals ();
+#endif
 
-  /* Socket activation -- we are handling connections on pre-opened
-   * file descriptors [FIRST_SOCKET_ACTIVATION_FD ..
-   * FIRST_SOCKET_ACTIVATION_FD+nr_socks-1].
-   */
-  if (socket_activation) {
-    nr_socks = socket_activation;
-    debug ("using socket activation, nr_socks = %zu", nr_socks);
-    socks = malloc (sizeof (int) * nr_socks);
-    if (socks == NULL) {
-      perror ("malloc");
+  /* Lock the process into memory if requested. */
+  if (swap) {
+#ifdef HAVE_MLOCKALL
+    if (mlockall (MCL_CURRENT | MCL_FUTURE) == -1) {
+      fprintf (stderr, "%s: --swap: mlockall: %m\n", program_name);
       exit (EXIT_FAILURE);
     }
-    for (i = 0; i < nr_socks; ++i)
-      socks[i] = FIRST_SOCKET_ACTIVATION_FD + i;
+    debug ("mlockall done");
+#else
+    fprintf (stderr, "%s: mlockall (--swap option) "
+             "is not supported on this platform\n", program_name);
+    exit (EXIT_FAILURE);
+#endif
+  }
+
+  /* Socket activation: the ‘socket_activation’ variable (> 0) is the
+   * number of file descriptors from FIRST_SOCKET_ACTIVATION_FD to
+   * FIRST_SOCKET_ACTIVATION_FD+socket_activation-1.
+   */
+  if (socket_activation) {
+      if (sockets_reserve (&socks, socket_activation) == -1) {
+        perror ("realloc");
+        exit (EXIT_FAILURE);
+      }
+    for (i = 0; i < socket_activation; ++i) {
+      int s = FIRST_SOCKET_ACTIVATION_FD + i, r;
+      /* This can't fail because of the reservation above. */
+      r = sockets_append (&socks, s);
+      assert (r == 0);
+    }
+    debug ("using socket activation, nr_socks = %zu", socks.size);
     change_user ();
     write_pidfile ();
-    accept_incoming_connections (socks, nr_socks);
-    free_listening_sockets (socks, nr_socks); /* also closes them */
+    accept_incoming_connections (&socks);
     return;
   }
 
@@ -848,23 +963,25 @@ start_serving (void)
     change_user ();
     write_pidfile ();
     threadlocal_new_server_thread ();
-    if (handle_single_connection (0, 1) == -1)
-      exit (EXIT_FAILURE);
+    handle_single_connection (saved_stdin, saved_stdout);
     return;
   }
 
-  /* Handling multiple connections on TCP/IP or a Unix domain socket. */
+  /* Handling multiple connections on TCP/IP, Unix domain socket or
+   * AF_VSOCK.
+   */
   if (unixsocket)
-    socks = bind_unix_socket (&nr_socks);
+    bind_unix_socket (&socks);
+  else if (vsock)
+    bind_vsock (&socks);
   else
-    socks = bind_tcpip_socket (&nr_socks);
+    bind_tcpip_socket (&socks);
 
   run_command ();
   change_user ();
   fork_into_background ();
   write_pidfile ();
-  accept_incoming_connections (socks, nr_socks);
-  free_listening_sockets (socks, nr_socks);
+  accept_incoming_connections (&socks);
 }
 
 static void
