@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -42,6 +42,7 @@
 
 #include <guestfs.h>
 
+#include "cleanup.h"
 #include "web-server.h"
 
 #include "test.h"
@@ -50,10 +51,10 @@ int
 main (int argc, char *argv[])
 {
   const char *sockpath;
-  size_t n;
   guestfs_h *g;
   int r;
   char *data;
+  CLEANUP_FREE char *usp_param = NULL;
 
 #ifndef HAVE_CURLOPT_UNIX_SOCKET_PATH
   fprintf (stderr, "%s: curl does not support CURLOPT_UNIX_SOCKET_PATH\n",
@@ -68,13 +69,15 @@ main (int argc, char *argv[])
   }
 
   /* Start nbdkit. */
-  n = strlen ("unix-socket-path") + 1 + strlen (sockpath) + 1;
-  char usp_param[n];
-  snprintf (usp_param, n, "%s=%s", "unix-socket-path", sockpath);
+  if (asprintf (&usp_param, "unix-socket-path=%s", sockpath) == -1) {
+    perror ("asprintf");
+    exit (EXIT_FAILURE);
+  }
   if (test_start_nbdkit ("curl",
+                         "-D", "curl.verbose=1",
+                         "http://localhost/disk",
                          "cookie=foo=bar; baz=1;",
                          usp_param, /* unix-socket-path=... */
-                         "http://localhost/disk",
                          NULL) == -1)
     exit (EXIT_FAILURE);
 
