@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -35,6 +35,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
+#undef NDEBUG /* Keep test strong even for nbdkit built without assertions */
+#include <assert.h>
 
 #include <nbdkit-filter.h>
 
@@ -43,6 +45,12 @@
 #define str(s) #s
 #define DEBUG_FUNCTION nbdkit_debug ("%s: %s", layer, __func__)
 
+/* Perform sanity checking on next_ops(nxdata) stability */
+struct handle {
+  void *nxdata;
+  struct nbdkit_next_ops *next_ops;
+};
+
 static void
 test_layers_filter_load (void)
 {
@@ -57,7 +65,7 @@ test_layers_filter_unload (void)
 
 static int
 test_layers_filter_config (nbdkit_next_config *next, void *nxdata,
-                            const char *key, const char *value)
+                           const char *key, const char *value)
 {
   DEBUG_FUNCTION;
   return next (nxdata, key, value);
@@ -65,7 +73,7 @@ test_layers_filter_config (nbdkit_next_config *next, void *nxdata,
 
 static int
 test_layers_filter_config_complete (nbdkit_next_config_complete *next,
-                                     void *nxdata)
+                                    void *nxdata)
 {
   DEBUG_FUNCTION;
   return next (nxdata);
@@ -74,94 +82,148 @@ test_layers_filter_config_complete (nbdkit_next_config_complete *next,
 #define test_layers_filter_config_help          \
   "test_layers_" layer "_config_help"
 
+static int
+test_layers_filter_get_ready (nbdkit_next_get_ready *next,
+                              void *nxdata)
+{
+  DEBUG_FUNCTION;
+  return next (nxdata);
+}
+
+static int
+test_layers_filter_preconnect (nbdkit_next_preconnect *next,
+                               void *nxdata, int readonly)
+{
+  DEBUG_FUNCTION;
+  return next (nxdata, readonly);
+}
+
 static void *
 test_layers_filter_open (nbdkit_next_open *next, void *nxdata, int readonly)
 {
-  static int handle;
+  struct handle *h = malloc (sizeof *h);
 
-  DEBUG_FUNCTION;
+  if (!h) {
+    perror ("malloc");
+    exit (1);
+  }
+  h->nxdata = nxdata;
+  h->next_ops = NULL;
 
   if (next (nxdata, readonly) == -1)
     return NULL;
 
-  return &handle;
+  /* Debug after recursing, to show opposite order from .close */
+  DEBUG_FUNCTION;
+
+  return h;
 }
 
 static void
 test_layers_filter_close (void *handle)
 {
   DEBUG_FUNCTION;
+  free (handle);
 }
 
 static int
 test_layers_filter_prepare (struct nbdkit_next_ops *next_ops, void *nxdata,
-                             void *handle)
+                            void *handle, int readonly)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == NULL);
+  assert (h->nxdata == nxdata);
+  h->next_ops = next_ops;
   DEBUG_FUNCTION;
   return 0;
 }
 
 static int
 test_layers_filter_finalize (struct nbdkit_next_ops *next_ops, void *nxdata,
-                              void *handle)
+                             void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return 0;
 }
 
 static int64_t
 test_layers_filter_get_size (struct nbdkit_next_ops *next_ops, void *nxdata,
-                              void *handle)
+                             void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->get_size (nxdata);
 }
 
 static int
 test_layers_filter_can_write (struct nbdkit_next_ops *next_ops, void *nxdata,
-                               void *handle)
+                              void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_write (nxdata);
 }
 
 static int
 test_layers_filter_can_flush (struct nbdkit_next_ops *next_ops, void *nxdata,
-                               void *handle)
+                              void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_flush (nxdata);
 }
 
 static int
 test_layers_filter_is_rotational (struct nbdkit_next_ops *next_ops,
-                                   void *nxdata,
-                                   void *handle)
+                                  void *nxdata,
+                                  void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->is_rotational (nxdata);
 }
 
 static int
 test_layers_filter_can_trim (struct nbdkit_next_ops *next_ops, void *nxdata,
-                              void *handle)
+                             void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_trim (nxdata);
 }
 
 static int
 test_layers_filter_can_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
-                              void *handle)
+                             void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_zero (nxdata);
 }
 
 static int
 test_layers_filter_can_fua (struct nbdkit_next_ops *next_ops, void *nxdata,
-                             void *handle)
+                            void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_fua (nxdata);
 }
@@ -171,6 +233,9 @@ test_layers_filter_can_multi_conn (struct nbdkit_next_ops *next_ops,
                                    void *nxdata,
                                    void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_multi_conn (nxdata);
 }
@@ -180,53 +245,83 @@ test_layers_filter_can_extents (struct nbdkit_next_ops *next_ops,
                                 void *nxdata,
                                 void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->can_extents (nxdata);
 }
 
 static int
-test_layers_filter_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
-                           void *handle, void *buf,
-                           uint32_t count, uint64_t offset,
-                           uint32_t flags, int *err)
+test_layers_filter_can_cache (struct nbdkit_next_ops *next_ops,
+                              void *nxdata,
+                              void *handle)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
+  DEBUG_FUNCTION;
+  return next_ops->can_cache (nxdata);
+}
+
+static int
+test_layers_filter_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
+                          void *handle, void *buf,
+                          uint32_t count, uint64_t offset,
+                          uint32_t flags, int *err)
+{
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->pread (nxdata, buf, count, offset, flags, err);
 }
 
 static int
 test_layers_filter_pwrite (struct nbdkit_next_ops *next_ops, void *nxdata,
-                            void *handle,
-                            const void *buf, uint32_t count, uint64_t offset,
-                            uint32_t flags, int *err)
+                           void *handle,
+                           const void *buf, uint32_t count, uint64_t offset,
+                           uint32_t flags, int *err)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->pwrite (nxdata, buf, count, offset, flags, err);
 }
 
 static int
 test_layers_filter_flush (struct nbdkit_next_ops *next_ops, void *nxdata,
-                           void *handle,
-                           uint32_t flags, int *err)
+                          void *handle,
+                          uint32_t flags, int *err)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->flush (nxdata, flags, err);
 }
 
 static int
 test_layers_filter_trim (struct nbdkit_next_ops *next_ops, void *nxdata,
-                          void *handle, uint32_t count, uint64_t offset,
-                          uint32_t flags, int *err)
+                         void *handle, uint32_t count, uint64_t offset,
+                         uint32_t flags, int *err)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->trim (nxdata, count, offset, flags, err);
 }
 
 static int
 test_layers_filter_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
-                          void *handle, uint32_t count, uint64_t offset,
-                          uint32_t flags, int *err)
+                         void *handle, uint32_t count, uint64_t offset,
+                         uint32_t flags, int *err)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->zero (nxdata, count, offset, flags, err);
 }
@@ -237,18 +332,34 @@ test_layers_filter_extents (struct nbdkit_next_ops *next_ops, void *nxdata,
                             uint32_t flags, struct nbdkit_extents *extents,
                             int *err)
 {
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
   DEBUG_FUNCTION;
   return next_ops->extents (nxdata, count, offset, flags, extents, err);
 }
 
+static int
+test_layers_filter_cache (struct nbdkit_next_ops *next_ops, void *nxdata,
+                          void *handle, uint32_t count, uint64_t offset,
+                          uint32_t flags, int *err)
+{
+  struct handle *h = handle;
+
+  assert (h->next_ops == next_ops && h->nxdata == nxdata);
+  DEBUG_FUNCTION;
+  return next_ops->cache (nxdata, count, offset, flags, err);
+}
+
 static struct nbdkit_filter filter = {
   .name              = "testlayers" layer,
-  .version           = PACKAGE_VERSION,
   .load              = test_layers_filter_load,
   .unload            = test_layers_filter_unload,
   .config            = test_layers_filter_config,
   .config_complete   = test_layers_filter_config_complete,
   .config_help       = test_layers_filter_config_help,
+  .get_ready         = test_layers_filter_get_ready,
+  .preconnect        = test_layers_filter_preconnect,
   .open              = test_layers_filter_open,
   .close             = test_layers_filter_close,
   .prepare           = test_layers_filter_prepare,
@@ -262,12 +373,14 @@ static struct nbdkit_filter filter = {
   .can_fua           = test_layers_filter_can_fua,
   .can_multi_conn    = test_layers_filter_can_multi_conn,
   .can_extents       = test_layers_filter_can_extents,
+  .can_cache         = test_layers_filter_can_cache,
   .pread             = test_layers_filter_pread,
   .pwrite            = test_layers_filter_pwrite,
   .flush             = test_layers_filter_flush,
   .trim              = test_layers_filter_trim,
   .zero              = test_layers_filter_zero,
   .extents           = test_layers_filter_extents,
+  .cache             = test_layers_filter_cache,
 };
 
 NBDKIT_REGISTER_FILTER(filter)
