@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018-2019 Red Hat Inc.
+ * Copyright (C) 2018-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -288,6 +288,32 @@ main (int argc, char *argv[])
      "test_layers_plugin_config_complete",
      NULL);
 
+  /* get_ready methods called in order. */
+  log_verify_seen_in_order
+    ("testlayersfilter3: get_ready",
+     "filter3: test_layers_filter_get_ready",
+     "testlayersfilter2: get_ready",
+     "filter2: test_layers_filter_get_ready",
+     "testlayersfilter1: get_ready",
+     "filter1: test_layers_filter_get_ready",
+     "testlayersplugin: get_ready",
+     "test_layers_plugin_get_ready",
+     NULL);
+
+  /* preconnect methods called in outer-to-inner order, complete
+   * in inner-to-outer order.
+   */
+  log_verify_seen_in_order
+    ("testlayersfilter3: preconnect",
+     "filter3: test_layers_filter_preconnect",
+     "testlayersfilter2: preconnect",
+     "filter2: test_layers_filter_preconnect",
+     "testlayersfilter1: preconnect",
+     "filter1: test_layers_filter_preconnect",
+     "testlayersplugin: preconnect",
+     "test_layers_plugin_preconnect",
+     NULL);
+
   /* open methods called in outer-to-inner order, but thanks to next
    * pointer, complete in inner-to-outer order. */
   log_verify_seen_in_order
