@@ -1,6 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
-# Copyright (C) 2018-2019 Red Hat Inc.
+# Copyright (C) 2018-2020 Red Hat Inc.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -48,7 +48,7 @@ start_t=$SECONDS
 nbdkit -v -U - \
        sh - \
        --filter=retry retry-delay=1 \
-       --run 'nbdsh --uri $uri -c "
+       --run 'nbdsh --uri "$uri" -c "
 h.zero (512, 0)
 try:
     h.zero (512, 0,
