@@ -1,6 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
-# Copyright (C) 2018-2019 Red Hat Inc.
+# Copyright (C) 2018-2020 Red Hat Inc.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -38,6 +38,7 @@ requires test "x$vddkdir" != "x"
 requires test -d "$vddkdir"
 requires test -f "$vddkdir/lib64/libvixDiskLib.so"
 requires qemu-img --version
+requires stat --version
 
 # VDDK > 5.1.1 only supports x86_64.
 if [ `uname -m` != "x86_64" ]; then
@@ -45,21 +46,31 @@ if [ `uname -m` != "x86_64" ]; then
     exit 77
 fi
 
-files="test-vddk-real.vmdk test-vddk-real.out"
+files="test-vddk-real.vmdk test-vddk-real.out test-vddk-real.log"
 rm -f $files
 cleanup_fn rm -f $files
 
 qemu-img create -f vmdk test-vddk-real.vmdk 100M
 
-export old_ld_library_path="$LD_LIBRARY_PATH"
-export LD_LIBRARY_PATH="$vddkdir/lib64:$LD_LIBRARY_PATH"
+# Since we are comparing error messages below, let's make sure we're
+# not translating errors.
+export LANG=C
 
+fail=0
 nbdkit -f -v -U - \
        --filter=readahead \
-       vddk libdir="$vddkdir" file=test-vddk-real.vmdk \
-       --run '
-       # VDDK library path breaks qemu-img, we must restore the
-       # original path here.
-       export LD_LIBRARY_PATH="$old_ld_library_path"
-       qemu-img convert $nbd -O raw test-vddk-real.out
-'
+       vddk libdir="$vddkdir" test-vddk-real.vmdk \
+       --run 'qemu-img convert $nbd -O raw test-vddk-real.out' \
+       > test-vddk-real.log 2>&1 || fail=1
+
+# Check the log for missing modules
+cat test-vddk-real.log
+if grep 'cannot open shared object file' test-vddk-real.log; then
+   exit 1
+fi
+
+# Check the raw output file has exactly the right size.
+size="$(stat -c '%s' test-vddk-real.out)"
+test "$size" -eq $((100 * 1024 * 1024))
+
+exit $fail
