@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2018 Red Hat Inc.
+ * Copyright (C) 2017-2020 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -72,6 +72,7 @@
 #include <time.h>
 
 #include "options.h"
+#include "utils.h"
 
 /* Construct an array of parameters passed through to real nbdkit. */
 static const char **cmd;
@@ -122,10 +123,12 @@ print_command (void)
   size_t i;
 
   if (len > 0)
-    fprintf (stderr, "%s", cmd[0]);
-  for (i = 1; i < len && cmd[i] != NULL; ++i)
-    fprintf (stderr, " %s", cmd[i]);
-  fprintf (stderr, "\n");
+    shell_quote (cmd[0], stderr);
+  for (i = 1; i < len && cmd[i] != NULL; ++i) {
+    fputc (' ', stderr);
+    shell_quote (cmd[i], stderr);
+  }
+  fputc ('\n', stderr);
 }
 
 int
@@ -136,6 +139,7 @@ main (int argc, char *argv[])
   time_t t;
   unsigned tu;
   char ts[32];
+  int r;
 
   /* If NBDKIT_VALGRIND=1 is set in the environment, then we run the
    * program under valgrind.  This is used by the tests.  Similarly if
@@ -164,6 +168,30 @@ main (int argc, char *argv[])
     }
   }
 
+  /* Needed for plugins written in OCaml. */
+  s = getenv ("LD_LIBRARY_PATH");
+  if (s)
+    r = asprintf (&s, "%s/plugins/ocaml/.libs:%s", builddir, s);
+  else
+    r = asprintf (&s, "%s/plugins/ocaml/.libs", builddir);
+  if (r < 0) {
+    perror ("asprintf");
+    exit (EXIT_FAILURE);
+  }
+  setenv ("LD_LIBRARY_PATH", s, 1);
+  free (s);
+  s = getenv ("LIBRARY_PATH");
+  if (s)
+    r = asprintf (&s, "%s/plugins/ocaml/.libs:%s", builddir, s);
+  else
+    r = asprintf (&s, "%s/plugins/ocaml/.libs", builddir);
+  if (r < 0) {
+    perror ("asprintf");
+    exit (EXIT_FAILURE);
+  }
+  setenv ("LIBRARY_PATH", s, 1);
+  free (s);
+
   /* Absolute path of the real nbdkit command. */
   passthru_format ("%s/server/nbdkit", builddir);
 
